import logging
import time
import traceback
from struct import pack, unpack
from threading import Thread

from pylgbst.messages import (
    MsgHubProperties,
    MsgPortOutput,
    MsgPortInputFmtSetupSingle,
    MsgPortInfoRequest,
    MsgPortModeInfoRequest,
    MsgPortInfo,
    MsgPortModeInfo,
    MsgPortInputFmtSingle,
)
from pylgbst.utilities import queue, str2hex, usbyte, ushort, usint, abs_scaled_100

log = logging.getLogger("peripherals")

# COLORS
COLOR_BLACK = 0x00
COLOR_PINK = 0x01
COLOR_PURPLE = 0x02
COLOR_BLUE = 0x03
COLOR_LIGHTBLUE = 0x04
COLOR_CYAN = 0x05
COLOR_GREEN = 0x06
COLOR_YELLOW = 0x07
COLOR_ORANGE = 0x08
COLOR_RED = 0x09
COLOR_WHITE = 0x0A
COLOR_NONE = 0xFF
COLORS = {
    COLOR_BLACK: "BLACK",
    COLOR_PINK: "PINK",
    COLOR_PURPLE: "PURPLE",
    COLOR_BLUE: "BLUE",
    COLOR_LIGHTBLUE: "LIGHTBLUE",
    COLOR_CYAN: "CYAN",
    COLOR_GREEN: "GREEN",
    COLOR_YELLOW: "YELLOW",
    COLOR_ORANGE: "ORANGE",
    COLOR_RED: "RED",
    COLOR_WHITE: "WHITE",
    COLOR_NONE: "NONE",
}


# TODO: support more types of peripherals from
# https://lego.github.io/lego-ble-wireless-protocol-docs/index.html#io-type-id


class Peripheral:
    """
    :type parent: pylgbst.hub.Hub
    :type _incoming_port_data: queue.Queue
    :type _port_mode: MsgPortInputFmtSingle
    """

    def __init__(self, parent, port):
        """
        :type parent: pylgbst.hub.Hub
        :type port: int
        """
        self.virtual_ports = ()
        self.hub = parent
        self.port = port

        self.is_buffered = False

        self._subscribers = set()
        self._port_mode = MsgPortInputFmtSingle(self.port, None, False, 1)

        self._incoming_port_data = queue.Queue(1)  # limit 1 means we drop data if we can't handle it fast enough
        thr = Thread(target=self._queue_reader)
        thr.daemon = True
        thr.name = "Port data queue: %s" % self
        thr.start()

    def __repr__(self):
        msg = "%s on port 0x%x" % (self.__class__.__name__, self.port)
        if self.virtual_ports:
            msg += " (ports 0x%x and 0x%x combined)" % (self.virtual_ports[0], self.virtual_ports[1])
        return msg

    def set_port_mode(self, mode, send_updates=None, update_delta=None):
        assert not self.virtual_ports, "TODO: support combined mode for sensors"

        if send_updates is None:
            send_updates = self._port_mode.upd_enabled
            log.debug("Implied update is enabled=%s", send_updates)

        if update_delta is None:
            update_delta = self._port_mode.upd_delta
            log.debug("Implied update delta=%s", update_delta)

        if (
                self._port_mode.mode == mode
                and self._port_mode.upd_enabled == send_updates
                and self._port_mode.upd_delta == update_delta
        ):
            log.debug("Already in target mode, no need to switch")
            return
        else:
            msg = MsgPortInputFmtSetupSingle(self.port, mode, update_delta, send_updates)
            resp = self.hub.send(msg)
            assert isinstance(resp, MsgPortInputFmtSingle)
            self._port_mode = resp

    def _send_output(self, msg):
        assert isinstance(msg, MsgPortOutput)
        msg.is_buffered = self.is_buffered  # TODO: support buffering
        self.hub.send(msg)

    def get_sensor_data(self, mode):
        self.set_port_mode(mode)
        msg = MsgPortInfoRequest(self.port, MsgPortInfoRequest.INFO_PORT_VALUE)
        resp = self.hub.send(msg)
        return self._decode_port_data(resp)

    def subscribe(self, callback, mode=0x00, granularity=1):
        if self._port_mode.mode != mode and self._subscribers:
            raise ValueError(
                "Port is in active mode %r, unsubscribe all subscribers first"
                % self._port_mode
            )
        self.set_port_mode(mode, True, granularity)
        if callback:
            self._subscribers.add(callback)

    def unsubscribe(self, callback=None):
        if callback in self._subscribers:
            self._subscribers.remove(callback)

        if not self._port_mode.upd_enabled:
            log.warning("Attempt to unsubscribe while port value updates are off: %s", self)
        elif not self._subscribers:
            self.set_port_mode(self._port_mode.mode, False)

    def _notify_subscribers(self, *args, **kwargs):
        for subscriber in self._subscribers.copy():
            subscriber(*args, **kwargs)
        return args

    def queue_port_data(self, msg):
        try:
            self._incoming_port_data.put_nowait(msg)
        except queue.Full:
            log.debug("Dropped port data: %r", msg)

    def _decode_port_data(self, msg):
        """Return the sensor value according to the current sensor mode
        :rtype: tuple
        """
        log.warning("Unhandled port data: %r", msg)
        return ()

    def _handle_port_data(self, msg):
        """
        :type msg: pylgbst.messages.MsgPortValueSingle
        """
        decoded = self._decode_port_data(msg)
        assert isinstance(decoded, (tuple, list)), "Unexpected data type: %s" % type(decoded)
        self._notify_subscribers(*decoded)

    def _queue_reader(self):
        while True:
            msg = self._incoming_port_data.get()
            try:
                self._handle_port_data(msg)
            except BaseException:
                log.warning("%s", traceback.format_exc())
                log.warning("Failed to handle port data by %s: %r", self, msg)

    def describe_possible_modes(self):
        mode_info = self.hub.send(MsgPortInfoRequest(self.port, MsgPortInfoRequest.INFO_MODE_INFO))
        assert isinstance(mode_info, MsgPortInfo)
        info = {
            "mode_count": mode_info.total_modes,
            "input_modes": [],
            "output_modes": [],
            "capabilities": {
                "logically_combinable": mode_info.is_combinable(),
                "synchronizable": mode_info.is_synchronizable(),
                "can_output": mode_info.is_output(),
                "can_input": mode_info.is_input(),
            },
        }

        if mode_info.is_combinable():
            mode_combinations = self.hub.send(
                MsgPortInfoRequest(self.port, MsgPortInfoRequest.INFO_MODE_COMBINATIONS)
            )
            assert isinstance(mode_combinations, MsgPortInfo)
            info['possible_mode_combinations'] = mode_combinations.possible_mode_combinations

        info["modes"] = []
        for mode in range(256):
            info["modes"].append(self._describe_mode(mode))

        for mode in mode_info.output_modes:
            info["output_modes"].append(self._describe_mode(mode))

        for mode in mode_info.input_modes:
            info["input_modes"].append(self._describe_mode(mode))

        log.debug("Port info for 0x%x: %s", self.port, info)
        return info

    def _describe_mode(self, mode):
        descr = {"Mode": mode}
        for info in MsgPortModeInfoRequest.INFO_TYPES:
            try:
                resp = self.hub.send(MsgPortModeInfoRequest(self.port, mode, info))
                assert isinstance(resp, MsgPortModeInfo)
                descr[MsgPortModeInfoRequest.INFO_TYPES[info]] = resp.value
            except RuntimeError:
                log.debug("Got error while requesting info 0x%x: %s", info, traceback.format_exc())
                if info == MsgPortModeInfoRequest.INFO_NAME:
                    break
        return descr


class LEDRGB(Peripheral):
    MODE_INDEX = 0x00
    MODE_RGB = 0x01

    def __init__(self, parent, port):
        super().__init__(parent, port)

    def set_color(self, color):
        """Set color of the RGB LED

        :param color: Accept 2 types of data:
            - RGB: tuple of 3 values,
            - index: 1 value choosen among 12 color values:
            `COLOR_BLACK, COLOR_PINK, COLOR_PURPLE, COLOR_BLUE, COLOR_LIGHTBLUE,
            COLOR_CYAN, COLOR_GREEN, COLOR_YELLOW, COLOR_ORANGE, COLOR_RED,
            COLOR_WHITE, COLOR_NONE`.
            Note that `COLOR_BLACK` and `COLOR_NONE` turn LED off.
        :type color: <tuple> or <int> or constant
        """
        if isinstance(color, (list, tuple)):
            assert len(color) == 3, "RGB color has to have 3 values"
            self.set_port_mode(self.MODE_RGB)
            payload = (
                    pack("<B", self.MODE_RGB)
                    + pack("<B", color[0])
                    + pack("<B", color[1])
                    + pack("<B", color[2])
            )
        else:
            if color == COLOR_NONE:
                color = COLOR_BLACK

            if color not in COLORS:
                raise ValueError("Color %s is not in list of available colors" % color)

            self.set_port_mode(self.MODE_INDEX)
            payload = pack("<B", self.MODE_INDEX) + pack("<B", color)

        msg = MsgPortOutput(self.port, MsgPortOutput.WRITE_DIRECT_MODE_DATA, payload)
        self._send_output(msg)

    def _decode_port_data(self, msg):
        """Decode data emitted by the hub

        .. note:: Experimental function for now; Most of the time, hardware
            shouldn't return any data. Also note that its seems not possible
            to retrieve the current RGB color on the peripheral.
            See issue #111.

        :param msg: Message retrieved from the hub.
        :return: Tuple of length dependent of the current mode (1 for MODE_INDEX,
            3 for MODE_RGB)
        """
        if len(msg.payload) == 3:
            return (
                usbyte(msg.payload, 0),
                usbyte(msg.payload, 1),
                usbyte(msg.payload, 2),
            )
        else:
            return usbyte(msg.payload, 0),

    # Set color of the RGB LED (no getter)
    color = property(fset=set_color)


class LEDLight(Peripheral):
    """Support of headlight kit (LPF2-LIGHT)"""

    MODE_BRIGHTNESS = 0x00

    def __init__(self, parent, port):
        super().__init__(parent, port)

    def set_brightness(self, brightness):
        """Set brightness of LEDs

        :param brightness: Number between 0 and 100%.
        :type brightness: <int> or <float>
        """
        if (
                not isinstance(brightness, (int, float))
                or brightness > 100
                or brightness < 0
        ):
            raise ValueError("Brightness must be a number between 0 and 100")

        self.set_port_mode(self.MODE_BRIGHTNESS)
        payload = pack("<B", self.MODE_BRIGHTNESS) + pack("<B", int(brightness))

        msg = MsgPortOutput(self.port, MsgPortOutput.WRITE_DIRECT_MODE_DATA, payload)
        self._send_output(msg)

    @property
    def brightness(self):
        """Get current brightness value
        :rtype: <int>
        """
        return self.get_sensor_data(LEDLight.MODE_BRIGHTNESS)[0]

    @brightness.setter
    def brightness(self, value):
        """Set brightness of LEDs
        .. seealso:: `set_brightness`
        """
        self.set_brightness(value)

    def _decode_port_data(self, msg):
        return usbyte(msg.payload, 0),


class BaseMotor(Peripheral):
    def _write_direct_mode(self, subcmd, params):
        params = pack("<B", subcmd) + params
        msg = MsgPortOutput(self.port, MsgPortOutput.WRITE_DIRECT_MODE_DATA, params)
        self._send_output(msg)


class TrainMotor(BaseMotor):
    """
    Simple DC motor (Lego 88011).
    See https://github.com/undera/pylgbst/issues/129
    """
    SUBCMD_POWER = 0x00
    SUBCMD_1 = 0x01  # TODO: figure out what it does. We know it's not sensor mode.

    def power(self, param=1.0):
        """
        Power the motor, with value -1.0..1.0
        """
        params = pack("<b", abs_scaled_100(param))
        self._write_direct_mode(self.SUBCMD_POWER, params)

    def stop(self):
        self.power(0)


class Motor(BaseMotor):
    SUBCMD_START_POWER = 0x01
    SUBCMD_START_POWER_GROUPED = 0x02
    SUBCMD_SET_ACC_TIME = 0x05
    SUBCMD_SET_DEC_TIME = 0x06
    SUBCMD_START_SPEED = 0x07
    # SUBCMD_START_SPEED = 0x08
    SUBCMD_START_SPEED_FOR_TIME = 0x09
    # SUBCMD_START_SPEED_FOR_TIME = 0x0A

    END_STATE_BRAKE = 127
    END_STATE_HOLD = 126
    END_STATE_FLOAT = 0

    def __init__(self, parent, port):
        super().__init__(parent, port)
        self.cmd_in_progress = False

    def _speed_abs(self, relative):  # FIXME: it's not "speed", rather it's a "power"
        if relative == Motor.END_STATE_BRAKE or relative == Motor.END_STATE_HOLD:
            # special value for BRAKE
            # https://lego.github.io/lego-ble-wireless-protocol-docs/index.html#output-sub-command-startpower-power
            return relative

        return abs_scaled_100(relative)

    def _send_cmd(self, subcmd, params, wait_complete=True):
        if self.virtual_ports:
            subcmd += 1  # de-facto rule

        msg = MsgPortOutput(self.port, subcmd, params, wait_complete)
        self._send_output(msg)

    def start_power(self, power_primary=1.0, power_secondary=None):
        """
        https://lego.github.io/lego-ble-wireless-protocol-docs/index.html#output-sub-command-startpower-power
        """
        if power_secondary is None:
            power_secondary = power_primary

        if self.virtual_ports:
            cmd = self.SUBCMD_START_POWER_GROUPED - 1  # because _send_cmd will do +1
        else:
            cmd = self.SUBCMD_START_POWER

        params = b""
        params += pack("<b", self._speed_abs(power_primary))
        if self.virtual_ports:
            params += pack("<b", self._speed_abs(power_secondary))

        self._send_cmd(cmd, params)

    def stop(self):
        self.timed(0)

    def set_acc_profile(self, seconds, profile_no=0x00):
        """
        https://lego.github.io/lego-ble-wireless-protocol-docs/index.html#output-sub-command-setacctime-time-profileno-0x05
        """
        params = b""
        params += pack("<H", int(seconds * 1000))
        params += pack("<B", profile_no)

        self._send_cmd(self.SUBCMD_SET_ACC_TIME, params)

    def set_dec_profile(self, seconds, profile_no=0x00):
        """
        https://lego.github.io/lego-ble-wireless-protocol-docs/index.html#output-sub-command-setdectime-time-profileno-0x06
        """
        params = b""
        params += pack("<H", int(seconds * 1000))
        params += pack("<B", profile_no)

        self._send_cmd(self.SUBCMD_SET_DEC_TIME, params)

    def start_speed(self, speed_primary=1.0, speed_secondary=None, max_power=1.0, use_profile=0b11):
        """
        https://lego.github.io/lego-ble-wireless-protocol-docs/index.html#output-sub-command-startspeed-speed-maxpower-useprofile-0x07
        """
        if speed_secondary is None:
            speed_secondary = speed_primary

        params = b""
        params += pack("<b", self._speed_abs(speed_primary))
        if self.virtual_ports:
            params += pack("<b", self._speed_abs(speed_secondary))

        params += pack("<B", int(100 * max_power))
        params += pack("<B", use_profile)

        self._send_cmd(self.SUBCMD_START_SPEED, params)

    def timed(self, seconds, speed_primary=1.0, speed_secondary=None, max_power=1.0, end_state=END_STATE_BRAKE,
              use_profile=0b11, wait_complete=True):
        """
        https://lego.github.io/lego-ble-wireless-protocol-docs/index.html#output-sub-command-startspeedfortime-time-speed-maxpower-endstate-useprofile-0x09
        """
        if speed_secondary is None:
            speed_secondary = speed_primary

        params = b""
        params += pack("<H", int(seconds * 1000))
        params += pack("<b", self._speed_abs(speed_primary))
        if self.virtual_ports:
            params += pack("<b", self._speed_abs(speed_secondary))

        params += pack("<B", int(100 * max_power))
        params += pack("<B", end_state)
        params += pack("<B", use_profile)

        self._send_cmd(self.SUBCMD_START_SPEED_FOR_TIME, params, wait_complete)

    def wait_complete(self):
        while self.cmd_in_progress:
            time.sleep(0.01)


class EncodedMotor(Motor):
    SUBCMD_START_SPEED_FOR_DEGREES = 0x0B
    # SUBCMD_START_SPEED_FOR_DEGREES = 0x0C
    SUBCMD_GOTO_ABSOLUTE_POSITION = 0x0D
    # SUBCMD_GOTO_ABSOLUTE_POSITIONC = 0x0E
    SUBCMD_PRESET_ENCODER = 0x14

    SENSOR_POWER = 0x00  # it's not input mode, hovewer returns some numbers
    SENSOR_SPEED = 0x01
    SENSOR_ANGLE = 0x02
    SENSOR_TEST = 0x03  # exists, but neither input nor output mode

    def angled(self, degrees, speed_primary=1.0, speed_secondary=None, max_power=1.0, end_state=Motor.END_STATE_BRAKE,
               use_profile=0b11, wait_complete=True):
        """
        https://lego.github.io/lego-ble-wireless-protocol-docs/index.html#output-sub-command-startspeedfordegrees-degrees-speed-maxpower-endstate-useprofile-0x0b
        :type degrees: int
        :type speed_primary: float
        """
        if speed_secondary is None:
            speed_secondary = speed_primary

        degrees = int(round(degrees))
        if degrees < 0:
            degrees = -degrees
            speed_primary = -speed_primary
            speed_secondary = -speed_secondary

        params = b""
        params += pack("<I", degrees)
        params += pack("<b", self._speed_abs(speed_primary))
        if self.virtual_ports:
            params += pack("<b", self._speed_abs(speed_secondary))

        params += pack("<B", int(100 * max_power))
        params += pack("<B", end_state)
        params += pack("<B", use_profile)

        self._send_cmd(self.SUBCMD_START_SPEED_FOR_DEGREES, params, wait_complete)

    def goto_position(self, degrees_primary, degrees_secondary=None, speed=1.0, max_power=1.0,
                      end_state=Motor.END_STATE_BRAKE, use_profile=0b11, wait_complete=True):
        """
        https://lego.github.io/lego-ble-wireless-protocol-docs/index.html#output-sub-command-gotoabsoluteposition-abspos-speed-maxpower-endstate-useprofile-0x0d
        """
        if degrees_secondary is None:
            degrees_secondary = degrees_primary

        params = b""
        params += pack("<i", degrees_primary)
        if self.virtual_ports:
            params += pack("<i", degrees_secondary)

        params += pack("<b", self._speed_abs(speed))

        params += pack("<B", int(100 * max_power))
        params += pack("<B", end_state)
        params += pack("<B", use_profile)

        self._send_cmd(self.SUBCMD_GOTO_ABSOLUTE_POSITION, params, wait_complete)

    def _decode_port_data(self, msg):
        data = msg.payload
        if self._port_mode.mode == self.SENSOR_ANGLE:
            angle = unpack("<l", data[0:4])[0]
            return angle,
        elif self._port_mode.mode == self.SENSOR_SPEED:
            speed = unpack("<b", data[0:1])[0]
            return speed,
        else:
            log.debug("Got motor sensor data while in unexpected mode: %r", self._port_mode)
            return ()

    def subscribe(self, callback, mode=SENSOR_ANGLE, granularity=1):
        super().subscribe(callback, mode, granularity)

    def preset_encoder(self, degrees=0, degrees_secondary=None, only_combined=False):
        """
        https://lego.github.io/lego-ble-wireless-protocol-docs/index.html#output-sub-command-presetencoder-position-n-a
        https://lego.github.io/lego-ble-wireless-protocol-docs/index.html#output-sub-command-presetencoder-leftposition-rightposition-0x14
        """
        if degrees_secondary is None:
            degrees_secondary = degrees

        if self.virtual_ports and not only_combined:
            self._send_cmd(self.SUBCMD_PRESET_ENCODER, pack("<i", degrees) + pack("<i", degrees_secondary))
        else:
            params = pack("<i", degrees)
            self._write_direct_mode(self.SENSOR_ANGLE, params)


class TiltSensor(Peripheral):
    MODE_2AXIS_ANGLE = 0x00
    MODE_2AXIS_SIMPLE = 0x01
    MODE_3AXIS_SIMPLE = 0x02
    MODE_IMPACT_COUNT = 0x03
    MODE_3AXIS_ACCEL = 0x04
    MODE_ORIENT_CF = 0x05
    MODE_IMPACT_CF = 0x06
    MODE_CALIBRATION = 0x07

    TRI_BACK = 0x00
    TRI_UP = 0x01
    TRI_DOWN = 0x02
    TRI_LEFT = 0x03
    TRI_RIGHT = 0x04
    TRI_FRONT = 0x05

    DUO_HORIZ = 0x00
    DUO_DOWN = 0x03
    DUO_LEFT = 0x05
    DUO_RIGHT = 0x07
    DUO_UP = 0x09

    DUO_STATES = {
        DUO_HORIZ: "HORIZONTAL",
        DUO_DOWN: "DOWN",
        DUO_LEFT: "LEFT",
        DUO_RIGHT: "RIGHT",
        DUO_UP: "UP",
    }

    TRI_STATES = {
        TRI_BACK: "BACK",
        TRI_UP: "UP",
        TRI_DOWN: "DOWN",
        TRI_LEFT: "LEFT",
        TRI_RIGHT: "RIGHT",
        TRI_FRONT: "FRONT",
    }

    def subscribe_accelerometer(self, callback, granularity=1):
        super().subscribe(callback, self.MODE_3AXIS_ACCEL, granularity)

    def subscribe(self, callback, mode=MODE_3AXIS_SIMPLE, granularity=1):
        super().subscribe(callback, mode, granularity)

    def _decode_port_data(self, msg):
        data = msg.payload
        if self._port_mode.mode == self.MODE_2AXIS_ANGLE:
            roll = unpack("<b", data[0:1])[0]
            pitch = unpack("<b", data[1:2])[0]
            return roll, pitch
        elif self._port_mode.mode == self.MODE_3AXIS_SIMPLE:
            state = usbyte(data, 0)
            return state,
        elif self._port_mode.mode == self.MODE_2AXIS_SIMPLE:
            state = usbyte(data, 0)
            return state,
        elif self._port_mode.mode == self.MODE_IMPACT_COUNT:
            bump_count = usint(data, 0)
            return bump_count,
        elif self._port_mode.mode == self.MODE_3AXIS_ACCEL:
            roll = unpack("<b", data[0:1])[0]
            pitch = unpack("<b", data[1:2])[0]
            yaw = unpack("<b", data[2:3])[0]  # did I get the order right?
            return roll, pitch, yaw
        elif self._port_mode.mode == self.MODE_ORIENT_CF:
            state = usbyte(data, 0)
            return state,
        elif self._port_mode.mode == self.MODE_IMPACT_CF:
            state = usbyte(data, 0)
            return state,
        elif self._port_mode.mode == self.MODE_CALIBRATION:
            return usbyte(data, 0), usbyte(data, 1), usbyte(data, 2)
        else:
            log.debug("Got tilt sensor data while in unexpected mode: %r", self._port_mode)
            return ()

    # TODO: add some methods from official doc, like
    # https://lego.github.io/lego-ble-wireless-protocol-docs/index.html#output-sub-command-tiltconfigimpact-impactthreshold-bumpholdoff-n-a


class VisionSensor(Peripheral):
    COLOR_INDEX = 0x00
    DISTANCE_INCHES = 0x01
    COUNT_2INCH = 0x02
    DISTANCE_REFLECTED = 0x03
    AMBIENT_LIGHT = 0x04
    SET_COLOR = 0x05
    COLOR_RGB = 0x06
    SET_IR_TX = 0x07
    COLOR_DISTANCE_FLOAT = 0x08

    DEBUG = 0x09  # first val is by fact ambient light, second is zero
    CALIBRATE = 0x0A  # gives constant values

    def __init__(self, parent, port):
        super().__init__(parent, port)

    def subscribe(self, callback, mode=COLOR_DISTANCE_FLOAT, granularity=1):
        super().subscribe(callback, mode, granularity)

    def _decode_port_data(self, msg):
        data = msg.payload
        if self._port_mode.mode == self.COLOR_INDEX:
            color = usbyte(data, 0)
            return color,
        elif self._port_mode.mode == self.COLOR_DISTANCE_FLOAT:
            color = usbyte(data, 0)
            val = usbyte(data, 1)
            partial = usbyte(data, 3)
            if partial:
                val += 1.0 / partial
            return color, float(val)
        elif self._port_mode.mode == self.DISTANCE_INCHES:
            val = usbyte(data, 0)
            return val,
        elif self._port_mode.mode == self.DISTANCE_REFLECTED:
            val = usbyte(data, 0) / 100.0
            return val,
        elif self._port_mode.mode == self.AMBIENT_LIGHT:
            val = usbyte(data, 0) / 100.0
            return val,
        elif self._port_mode.mode == self.COUNT_2INCH:
            count = usint(data, 0)
            return count,
        elif self._port_mode.mode == self.COLOR_RGB:
            val1 = int(255 * ushort(data, 0) / 1023.0)
            val2 = int(255 * ushort(data, 2) / 1023.0)
            val3 = int(255 * ushort(data, 4) / 1023.0)
            return val1, val2, val3
        elif self._port_mode.mode == self.DEBUG:
            val1 = 10 * ushort(data, 0) / 1023.0
            val2 = 10 * ushort(data, 2) / 1023.0
            return val1, val2
        elif self._port_mode.mode == self.CALIBRATE:
            return [ushort(data, x * 2) for x in range(8)]
        else:
            log.debug("Unhandled VisionSensor data in mode %s: %s", self._port_mode.mode, str2hex(data))
            return ()

    def set_color(self, color):
        """Set color of the RGB LED on the sensor

        :param color:
            Note that `COLOR_BLACK` and `COLOR_NONE` turn LED off.
        """
        if color == COLOR_NONE:
            color = COLOR_BLACK

        if color not in COLORS:
            raise ValueError("Color %s is not in list of available colors" % color)

        self.set_port_mode(self.SET_COLOR)
        payload = pack("<B", self.SET_COLOR) + pack("<B", color)

        msg = MsgPortOutput(self.port, MsgPortOutput.WRITE_DIRECT_MODE_DATA, payload)
        self._send_output(msg)

    def set_ir_tx(self, level=1.0):
        assert 0 <= level <= 1.0
        self.set_port_mode(self.SET_IR_TX)
        payload = pack("<B", self.SET_IR_TX) + pack("<H", int(level * 65535))

        msg = MsgPortOutput(self.port, MsgPortOutput.WRITE_DIRECT_MODE_DATA, payload)
        self._send_output(msg)

    @property
    def color(self):
        """Get index of the current color
        :return: See values of `COLORS` module variable.
        :rtype: <int>
        """
        return self.get_sensor_data(self.COLOR_INDEX)[0]

    @property
    def distance(self):
        """Get measured distance of an object to the sensor
        :return: Value from 0 to 10
        :rtype: <int>
        """
        return self.get_sensor_data(self.DISTANCE_INCHES)[0]

    @property
    def reflected_light(self):
        """Get measured reflected light
        :return: Value from 0 to 1.0
        :rtype: <float>
        """
        return self.get_sensor_data(self.DISTANCE_REFLECTED)[0]

    @property
    def luminosity(self):
        """Get the current luminosity in lux
        :return: Value from 0 to 1.0
        :rtype: <float>
        """
        return self.get_sensor_data(self.AMBIENT_LIGHT)[0]

    @property
    def detection_count(self):
        """Get the number of object detections ~2 inches in front of sensor
        :rtype: <int>
        """
        return self.get_sensor_data(self.COUNT_2INCH)[0]

    @property
    def rgb_color(self):
        """Get detected RGB channels

        :return: Tuple of 3 values for RGB channels
        :rtype: <tuple <int>, <int>, <int>>
        """
        return self.get_sensor_data(self.COLOR_RGB)


class Voltage(Peripheral):
    """Retrieve voltage information from the hub"""

    # sensor says there are "L" and "S" values, but what are they?
    VOLTAGE_L = 0x00
    VOLTAGE_S = 0x01

    def __init__(self, parent, port):
        super().__init__(parent, port)

    def _decode_port_data(self, msg):
        data = msg.payload
        val = ushort(data, 0)
        volts = 9600.0 * val / 3893.0 / 1000.0
        return volts,

    @property
    def voltage(self):
        """Get the measured voltage of the battery

        :return: Return the value (in Volts) of the VOLTAGE_L mode by default.
        :rtype: <float>
        """
        return self.get_sensor_data(self.VOLTAGE_L)[0]


class Current(Peripheral):
    """Retrieve current information from the hub"""

    CURRENT_L = 0x00
    CURRENT_S = 0x01

    def __init__(self, parent, port):
        super().__init__(parent, port)

    def _decode_port_data(self, msg):
        val = ushort(msg.payload, 0)
        milliampers = 2444 * val / 4095.0
        return milliampers,

    @property
    def current(self):
        """Get the measured current of the battery

        :return: Return the value (in mA) of the CURRENT_L mode by default.
        :rtype: <float>
        """
        return self.get_sensor_data(self.CURRENT_L)[0]


class Button(Peripheral):
    """
    It's not really a peripheral, we use MSG_DEVICE_INFO commands to interact with it
    """

    def __init__(self, parent):
        super().__init__(parent, 0)  # fake port 0
        self.hub.add_message_handler(MsgHubProperties, self._props_msg)

    def subscribe(self, callback, mode=None, granularity=1):
        self.hub.send(MsgHubProperties(MsgHubProperties.BUTTON, MsgHubProperties.UPD_ENABLE))

        if callback:
            self._subscribers.add(callback)

    def unsubscribe(self, callback=None):
        if callback in self._subscribers:
            self._subscribers.remove(callback)

        if not self._subscribers:
            self.hub.send(MsgHubProperties(MsgHubProperties.BUTTON, MsgHubProperties.UPD_DISABLE))

    def _props_msg(self, msg):
        """
        :type msg: MsgHubProperties
        """
        if (
                msg.property == MsgHubProperties.BUTTON
                and msg.operation == MsgHubProperties.UPSTREAM_UPDATE
        ):
            self._notify_subscribers(usbyte(msg.parameters, 0))


class RemoteButton(Peripheral):

    # this only works with modes rckey (0), keya (1), and keyr (2) as in
    # https://virantha.github.io/bricknil/lego_api/lego.html#remote-buttons

    RCKEY = 0x00
    KEYA = 0x01
    KEYR = 0x02

    button_sets = {0: "LEFT",
                   1: "RIGHT"}
    button_events = {b'\x01': "PLUS",
                     b'\x7f': "RED",
                     b'\xff': "MINUS",
                     b'\x00': "RELEASE"}

    def __init__(self, parent, port):
        super().__init__(parent, port)

        self.hub.add_message_handler(MsgHubProperties, self._props_msg)

    def subscribe(self, callback, mode=0x00, granularity=1):
        # override base class to prevent invalid mode
        if mode not in [self.RCKEY, self.KEYR, self.KEYA]:
            log.debug("Invalid mode: %i", mode)
            raise ValueError("Invalid mode: ", mode)

        super().subscribe(callback, mode)

    def _decode_port_data(self, msg):
        button = self.button_events[msg.payload]
        set = self.button_sets[msg.port]

        return (button, set)

    def _handle_port_data(self, msg):
        """
        :type msg: pylgbst.messages.MsgPortValueSingle
        """
        decoded = self._decode_port_data(msg)
        if decoded is not None:
            assert isinstance(decoded, (tuple, list)), "Unexpected data type: %s" % type(decoded)
            self._notify_subscribers(*decoded)

    def _props_msg(self, msg):
        """
        :type msg: MsgHubProperties
        """
        if (
                msg.property == MsgHubProperties.BUTTON
                and msg.operation == MsgHubProperties.UPSTREAM_UPDATE
        ):
            self._notify_subscribers(usbyte(msg.parameters, 0))


class Temperature(Peripheral):
    """Get battery temperature from the hub"""
    MODE_TEMP = 0

    def __init__(self, parent, port):
        super().__init__(parent, port)

    def _decode_port_data(self, msg):
        # Fix temp with a small offset to get the real temperature
        magic_offset = 2.1
        return (unpack("<h", msg.payload)[0] / 10) - magic_offset,

    @property
    def temperature(self):
        """Get the measured temperature in C°

        :rtype: <float>
        """
        return self.get_sensor_data(self.MODE_TEMP)[0]



import unittest

import pygatt
import serial
from pygatt import BLEAddressType
from pygatt.backends.bgapi.bgapi import MAX_CONNECTION_ATTEMPTS
from pygatt.backends.bgapi.device import BGAPIBLEDevice
from pygatt.backends.bgapi.util import USBSerialDeviceInfo

from pylgbst.comms.cpygatt import GattoolConnection
from tests import log


class SerialMock(serial.Serial):
    def write(self, data):
        self.is_open = True
        log.debug("Write data to serial: %s", data)
        return len(data)

    def flush(self, *args, **kwargs):
        pass

    def close(self):
        pass

    def read(self, size=1):
        return bytes()


class BGAPIBLEDeviceMock(BGAPIBLEDevice):
    def subscribe(self, uuid, callback=None, indication=False):
        log.debug("Mock subscribing")

    def char_write_handle(self, char_handle, value, wait_for_response=False):
        log.debug("Mock write: %s", value)


class BlueGigaBackendMock(pygatt.BGAPIBackend):
    def _open_serial_port(self, max_connection_attempts=MAX_CONNECTION_ATTEMPTS):
        log.debug("Mock open serial port")
        self._ser = SerialMock()

    def expect(self, expected, *args, **kargs):
        log.debug("Mock expect")
        data = {
            "packet_type": 0x04,
            "sender": "abcdef".encode('ascii'),
            "data": [1, 2, 3],
            "rssi": 1
        }
        self._ble_evt_gap_scan_response(data)

    def connect(self, address, timeout=5, address_type=BLEAddressType.public, interval_min=60, interval_max=76,
                supervision_timeout=100, latency=0):
        log.debug("Mock connect")
        device = BGAPIBLEDeviceMock("address", 0, self)
        return device

    def _detect_device_port(self):
        return USBSerialDeviceInfo().port_name


class BlueGigaTests(unittest.TestCase):
    def test_1(self):
        obj = GattoolConnection()
        obj.backend = BlueGigaBackendMock
        obj.connect(u'66:65:64:63:62:61')
        obj.write(0, "test".encode('ascii'))
        obj.set_notify_handler(lambda x: None)



import unittest

import pylgbst.comms.cbluepy as bp_backend


class PeripheralMock(object):
    def __init__(self, addr, addrType, ifaceNumber):
        pass

    def waitForNotifications(self, timeout):
        pass

    def writeCharacteristic(self, handle, data):
        pass

    def withDelegate(self, delegate):
        pass

    def disconnect(self):
        pass


bp_backend.PROPAGATE_DISPATCHER_EXCEPTION = True
bp_backend.btle.Peripheral = lambda *args, **kwargs: PeripheralMock(*args, **kwargs)


class BluepyTestCase(unittest.TestCase):
    def test_get_iface_number(self):
        self.assertEqual(bp_backend._get_iface_number('hci0'), 0)
        self.assertEqual(bp_backend._get_iface_number('hci10'), 10)
        try:
            bp_backend._get_iface_number('ads12')
            self.fail('Missing exception for incorrect value')
        except ValueError:
            pass

    def test_delegate(self):
        def _handler(handle, data):
            _handler.called = True

        delegate = bp_backend.BluepyDelegate(_handler)
        delegate.handleNotification(123, 'qwe')
        self.assertEqual(_handler.called, True)

    def test_threaded_peripheral(self):
        tp = bp_backend.BluepyThreadedPeripheral('address', 'addrType', 'hci0')
        self.assertEqual(tp._addr, 'address')
        self.assertEqual(tp._addrType, 'addrType')
        self.assertEqual(tp._iface_number, 0)
        self.assertNotEqual(tp._dispatcher_thread, None)

        # Schedule some methods to async queue and give them some time to resolve
        tp.set_notify_handler(lambda: '')
        tp.write(123, 'qwe')

        tp._dispatcher_thread.join(1)
        self.assertEqual(tp._dispatcher_thread.is_alive(), True)
        tp.disconnect()

        tp._dispatcher_thread.join(2)
        self.assertEqual(tp._dispatcher_thread.is_alive(), False)



import logging
import re
from threading import Thread, Event

from bluepy import btle

from pylgbst.comms import Connection
from pylgbst.utilities import str2hex, queue

log = logging.getLogger('comms-bluepy')

COMPLETE_LOCAL_NAME_ADTYPE = 9
PROPAGATE_DISPATCHER_EXCEPTION = False


def _get_iface_number(controller):
    """bluepy uses iface numbers instead of full names."""
    if not controller:
        return None
    m = re.search(r'hci(\d+)$', controller)
    if not m:
        raise ValueError('Cannot find iface number in {}.'.format(controller))
    return int(m.group(1))


class BluepyDelegate(btle.DefaultDelegate):
    def __init__(self, handler):
        btle.DefaultDelegate.__init__(self)

        self._handler = handler

    def handleNotification(self, cHandle, data):
        log.debug('Incoming notification')
        self._handler(cHandle, data)


# We need a separate thread to wait for notifications,
# but calling peripheral's methods from different threads creates issues,
# so we will wrap all the calls into a thread
class BluepyThreadedPeripheral:
    def __init__(self, addr, addrType, controller):
        self._call_queue = queue.Queue()
        self._addr = addr
        self._addrType = addrType
        self._iface_number = _get_iface_number(controller)

        self._disconnect_event = Event()

        self._dispatcher_thread = Thread(target=self._dispatch_calls)
        self._dispatcher_thread.setDaemon(True)
        self._dispatcher_thread.setName("Bluepy call dispatcher")
        self._dispatcher_thread.start()

    def _dispatch_calls(self):
        self._peripheral = btle.Peripheral(self._addr, self._addrType, self._iface_number)
        try:
            while not self._disconnect_event.is_set():
                try:
                    try:
                        method = self._call_queue.get(False)
                        method()
                    except queue.Empty:
                        pass
                    self._peripheral.waitForNotifications(1.)
                except Exception as ex:
                    log.exception('Exception in call dispatcher thread', exc_info=ex)
                    if PROPAGATE_DISPATCHER_EXCEPTION:
                        log.error("Terminating dispatcher thread.")
                        raise
        finally:
            self._peripheral.disconnect()

    def write(self, handle, data):
        self._call_queue.put(lambda: self._peripheral.writeCharacteristic(handle, data))

    def set_notify_handler(self, handler):
        delegate = BluepyDelegate(handler)
        self._call_queue.put(lambda: self._peripheral.withDelegate(delegate))

    def disconnect(self):
        self._disconnect_event.set()


class BluepyConnection(Connection):
    def __init__(self, controller='hci0'):
        Connection.__init__(self)
        self._peripheral = None  # :type BluepyThreadedPeripheral
        self._controller = controller

    def connect(self, hub_mac=None, hub_name=None):
        log.debug("Trying to connect client to MoveHub with MAC: %s", hub_mac)
        scanner = btle.Scanner()

        while not self._peripheral:
            log.info("Discovering devices...")
            scanner.scan(1)
            devices = scanner.getDevices()

            for dev in devices:
                address = dev.addr
                address_type = dev.addrType
                name = dev.getValueText(COMPLETE_LOCAL_NAME_ADTYPE)

                if self._is_device_matched(address, name, hub_mac, hub_name):
                    self._peripheral = BluepyThreadedPeripheral(address, address_type, self._controller)
                    break

        return self

    def disconnect(self):
        self._peripheral.disconnect()

    def write(self, handle, data):
        log.debug("Writing to handle %s: %s", handle, str2hex(data))
        self._peripheral.write(handle, data)

    def set_notify_handler(self, handler):
        self._peripheral.set_notify_handler(handler)

    def is_alive(self):
        return True



import time
import unittest

from pylgbst.hub import Hub, MoveHub
from pylgbst.messages import MsgHubAction, MsgHubAlert, MsgHubProperties
from pylgbst.peripherals import VisionSensor
from pylgbst.utilities import usbyte
from tests import ConnectionMock


class HubTest(unittest.TestCase):
    def test_hub_properties(self):
        conn = ConnectionMock().connect()
        hub = Hub(conn)

        conn.notification_delayed('060001060600', 0.1)
        msg = MsgHubProperties(MsgHubProperties.VOLTAGE_PERC, MsgHubProperties.UPD_REQUEST)
        resp = hub.send(msg)
        assert isinstance(resp, MsgHubProperties)
        self.assertEqual(1, len(resp.parameters))
        self.assertEqual(0, usbyte(resp.parameters, 0))

        conn.notification_delayed('12000101064c45474f204d6f766520487562', 0.1)
        msg = MsgHubProperties(MsgHubProperties.ADVERTISE_NAME, MsgHubProperties.UPD_REQUEST)
        resp = hub.send(msg)
        assert isinstance(resp, MsgHubProperties)
        self.assertEqual(b"LEGO Move Hub", resp.parameters)

        conn.wait_notifications_handled()

    def test_device_attached(self):
        conn = ConnectionMock().connect()
        hub = Hub(conn)

        # regular startup attaches
        conn.notifications.append('0f0004010126000000001000000010')
        conn.notifications.append('0f0004020125000000001000000010')

        # detach and reattach
        conn.notifications.append('0500040100')
        conn.notifications.append('0500040200')
        conn.notifications.append('0f0004010126000000001000000010')
        conn.notifications.append('0f0004020125000000001000000010')

        del hub
        conn.wait_notifications_handled()

    def test_hub_actions(self):
        conn = ConnectionMock().connect()
        hub = Hub(conn)
        conn.notification_delayed('04000230', 0.1)
        hub.send(MsgHubAction(MsgHubAction.UPSTREAM_SHUTDOWN))

        conn.notification_delayed('04000231', 0.1)
        hub.send(MsgHubAction(MsgHubAction.UPSTREAM_DISCONNECT))

        conn.notification_delayed('04000232', 0.1)
        hub.send(MsgHubAction(MsgHubAction.UPSTREAM_BOOT_MODE))
        time.sleep(0.1)
        conn.wait_notifications_handled()

    def test_hub_alerts(self):
        conn = ConnectionMock().connect()
        hub = Hub(conn)
        conn.notification_delayed('0600 03 0104ff', 0.1)
        hub.send(MsgHubAlert(MsgHubAlert.LOW_VOLTAGE, MsgHubAlert.UPD_REQUEST))
        conn.notification_delayed('0600 03 030400', 0.1)
        hub.send(MsgHubAlert(MsgHubAlert.LOW_SIGNAL, MsgHubAlert.UPD_REQUEST))
        conn.wait_notifications_handled()

    def test_error(self):
        conn = ConnectionMock().connect()
        hub = Hub(conn)
        conn.notification_delayed("0500056105", 0.1)
        time.sleep(0.2)
        conn.wait_notifications_handled()

    def test_disconnect_off(self):
        conn = ConnectionMock().connect()
        hub = Hub(conn)
        conn.notification_delayed("04000231", 0.1)
        hub.disconnect()
        self.assertEqual(b"04000202", conn.writes[1][1])

        conn = ConnectionMock().connect()
        hub = Hub(conn)
        conn.notification_delayed("04000230", 0.1)
        hub.switch_off()
        self.assertEqual(b"04000201", conn.writes[1][1])

    def test_sensor(self):
        conn = ConnectionMock().connect()
        conn.notifications.append("0f0004020125000000001000000010")  # add dev
        hub = Hub(conn)
        time.sleep(0.1)
        dev = hub.peripherals[0x02]
        assert isinstance(dev, VisionSensor)

        conn.notification_delayed("0a004702080000000000", 0.1)
        conn.notification_delayed("08004502ff0aff00", 0.2)  # value for sensor
        self.assertEqual((255, 10.0), dev.get_sensor_data(VisionSensor.COLOR_DISTANCE_FLOAT))
        self.assertEqual(b"0a004102080100000000", conn.writes.pop(1)[1])
        self.assertEqual(b"0500210200", conn.writes.pop(1)[1])

        vals = []
        cb = lambda x, y=None: vals.append((x, y))

        conn.notification_delayed("0a004702080100000001", 0.1)  # subscribe ack
        dev.subscribe(cb, granularity=1)
        self.assertEqual(b"0a004102080100000001", conn.writes.pop(1)[1])

        conn.notification_delayed("08004502ff0aff00", 0.1)  # value for sensor
        time.sleep(0.3)

        conn.notification_delayed("0a004702080000000000", 0.3)  # unsubscribe ack
        dev.unsubscribe(cb)
        self.assertEqual(b"0a004102080100000000", conn.writes.pop(1)[1])

        self.assertEqual([(255, 10.0)], vals)


class MoveHubTest(unittest.TestCase):
    def test_capabilities(self):
        conn = ConnectionMock()
        conn.notifications.append('0f00 04 02 0125000000001000000010')
        conn.notifications.append('0f00 04 03 0126000000001000000010')
        conn.notifications.append('0f00 04 00 0127000100000001000000')
        conn.notifications.append('0f00 04 01 0127000100000001000000')
        conn.notifications.append('0900 04 10 0227003738')
        conn.notifications.append('0f00 04 32 0117000100000001000000')
        conn.notifications.append('0f00 04 3a 0128000000000100000001')
        conn.notifications.append('0f00 04 3b 0115000200000002000000')
        conn.notifications.append('0f00 04 3c 0114000200000002000000')

        conn.notification_delayed('12000101064c45474f204d6f766520487562', 1.1)
        conn.notification_delayed('0b00010d06001653a0d1d4', 1.3)
        conn.notification_delayed('060001060600', 1.5)
        conn.notification_delayed('0600030104ff', 1.7)
        MoveHub(conn.connect())
        time.sleep(1)
        conn.wait_notifications_handled()
        self.assertEqual(b"0500010105", conn.writes[1][1])
        self.assertEqual(b"0500010d05", conn.writes[2][1])
        self.assertEqual(b"0500010605", conn.writes[3][1])
        self.assertEqual(b"0500030103", conn.writes[4][1])



"""
This package holds communication aspects
"""
import binascii
import json
import logging
import socket
import traceback
from abc import abstractmethod
from binascii import unhexlify
from threading import Thread

from pylgbst.messages import MsgHubAction
from pylgbst.utilities import str2hex

log = logging.getLogger('comms')

MOVE_HUB_HW_UUID_SERV = '00001623-1212-efde-1623-785feabcd123'
MOVE_HUB_HW_UUID_CHAR = '00001624-1212-efde-1623-785feabcd123'
ENABLE_NOTIFICATIONS_HANDLE = 0x000f
ENABLE_NOTIFICATIONS_VALUE = b'\x01\x00'

MOVE_HUB_HARDWARE_HANDLE = 0x0E


class Connection:
    def connect(self, hub_mac=None):
        pass

    @abstractmethod
    def is_alive(self):
        pass

    def disconnect(self):
        pass

    @abstractmethod
    def write(self, handle, data):
        pass

    @abstractmethod
    def set_notify_handler(self, handler):
        pass

    def enable_notifications(self):
        self.write(ENABLE_NOTIFICATIONS_HANDLE, ENABLE_NOTIFICATIONS_VALUE)

    def _is_device_matched(self, address, dev_name, hub_mac, find_name):
        assert hub_mac or find_name, 'You have to provide either hub_mac or hub_name in connection options'
        log.debug("Checking device: %s, MAC: %s", dev_name, address)
        matched = False
        if address != "00:00:00:00:00:00":
            if hub_mac:
                if hub_mac.lower() == address.lower():
                    matched = True
            elif dev_name == find_name:
                matched = True

            if matched:
                log.info("Found %s at %s", dev_name, address)

        return matched


class DebugServer:
    """
    WARNING: deprecated class
    Starts TCP server to be used with DebugServerConnection to speed-up development process
    It holds BLE connection to Move Hub, so no need to re-start it every time
    Usage: DebugServer(BLEConnection().connect()).start()

    :type connection: BLEConnection
    """

    def __init__(self, connection):
        self._running = False
        self.sock = socket.socket()
        self.connection = connection

    def start(self, port=9090):
        self.sock.bind(('', port))
        self.sock.listen(1)

        self._running = True
        while self._running:
            log.info("Accepting MoveHub debug connections at %s", port)
            conn, addr = self.sock.accept()
            if not self._running:
                raise KeyboardInterrupt("Shutdown")
            self.connection.set_notify_handler(lambda x, y: self._notify(conn, x, y))
            try:
                self._handle_conn(conn)
            except KeyboardInterrupt:
                raise
            except BaseException:
                log.error("Problem handling incoming connection: %s", traceback.format_exc())
            finally:
                self.connection.set_notify_handler(self._notify_dummy)
                conn.close()

    def __del__(self):
        self.sock.close()

    def _notify_dummy(self, handle, data):
        log.debug("Dropped notification from handle %s: %s", handle, binascii.hexlify(data))
        self._check_shutdown(data)

    def _notify(self, conn, handle, data):
        payload = {"type": "notification", "handle": handle, "data": str2hex(data)}
        log.debug("Send notification: %s", payload)
        try:
            conn.send(json.dumps(payload) + "\n")
        except KeyboardInterrupt:
            raise
        except BaseException:
            log.error("Problem sending notification: %s", traceback.format_exc())

        self._check_shutdown(data)

    def _check_shutdown(self, data):
        if data[5] == MsgHubAction.TYPE:
            log.warning("Device shutdown")
            self._running = False

    def _handle_conn(self, conn):
        """
        :type conn: socket._socketobject
        """
        buf = ""
        while True:
            data = conn.recv(1024)
            log.debug("Recv: %s", data.strip())
            if not data:
                break

            buf += data

            if "\n" in buf:
                line = buf[:buf.index("\n")]
                buf = buf[buf.index("\n") + 1:]

                if line:
                    log.debug("Cmd line: %s", line)
                    try:
                        self._handle_cmd(json.loads(line))
                    except KeyboardInterrupt:
                        raise
                    except BaseException:
                        log.error("Failed to handle cmd: %s", traceback.format_exc())

    def _handle_cmd(self, cmd):
        if cmd['type'] == 'write':
            self.connection.write(cmd['handle'], unhexlify(cmd['data']))
        else:
            raise ValueError("Unhandled cmd: %s", cmd)


class DebugServerConnection(Connection):
    """
    WARNING: deprecated class
    Connection type to be used with DebugServer, replaces BLEConnection
    """

    def __init__(self, port=9090):
        super().__init__()
        self.notify_handler = None
        self.buf = ""
        self.sock = socket.socket()
        self.sock.connect(('localhost', port))
        self.incoming = []

        self.reader = Thread(target=self._recv)
        self.reader.setName("Debug connection reader")
        self.reader.setDaemon(True)
        self.reader.start()

    def __del__(self):
        self.sock.close()

    def write(self, handle, data):
        payload = {
            "type": "write",
            "handle": handle,
            "data": str2hex(data)
        }
        self._send(payload)

    def _send(self, payload):
        log.debug("Sending to debug server: %s", payload)
        self.sock.send(json.dumps(payload) + "\n")

    def _recv(self):
        while True:
            data = self.sock.recv(1024)
            log.debug("Recv from debug server: %s", data.strip())
            if not data:
                raise KeyboardInterrupt("Server has closed connection")

            self.buf += data

            while "\n" in self.buf:
                line = self.buf[:self.buf.index("\n")]
                self.buf = self.buf[self.buf.index("\n") + 1:]
                if line:
                    item = json.loads(line)
                    if item['type'] == 'notification' and self.notify_handler:
                        try:
                            self.notify_handler(item['handle'], unhexlify(item['data']))
                        except BaseException:
                            log.error("Failed to notify handler: %s", traceback.format_exc())
                    elif item['type'] == 'response':
                        self.incoming.append(item)
                    else:
                        log.warning("Dropped inbound: %s", item)

    def set_notify_handler(self, handler):
        self.notify_handler = handler

    def is_alive(self):
        return self.reader.isAlive()



# noinspection PyMethodOverriding
import logging
import traceback
from threading import Thread

from gattlib import DiscoveryService, GATTRequester

from pylgbst.comms import Connection
from pylgbst.utilities import queue, str2hex

log = logging.getLogger('comms-gattlib')


class Requester(GATTRequester):
    """
    Wrapper to access `on_notification` capability of GATT
    Set "notification_sink" field to a callable that will handle incoming data
    """

    def __init__(self, p_object, *args, **kwargs):
        super().__init__(p_object, *args, **kwargs)
        self.notification_sink = None

        self._notify_queue = queue.Queue()  # this queue is to minimize time spent in gattlib C code
        self.notify_thread = Thread(target=self._dispatch_notifications)
        self.notify_thread.setDaemon(True)
        self.notify_thread.setName("Notify queue dispatcher")
        self.notify_thread.start()

    def on_notification(self, handle, data):
        # log.debug("requester notified, sink: %s", self.notification_sink)
        self._notify_queue.put((handle, data))

    def on_indication(self, handle, data):
        log.debug("Indication on handle %s: %s", handle, str2hex(data))

    def _dispatch_notifications(self):
        while True:
            handle, data = self._notify_queue.get()
            data = data[3:]  # for some reason, there are extra bytes
            if self.notification_sink:
                try:
                    self.notification_sink(handle, data)
                except BaseException:
                    log.warning("Data was: %s", str2hex(data))
                    log.warning("Failed to dispatch notification: %s", traceback.format_exc())
            else:
                log.warning("Dropped notification %s: %s", handle, str2hex(data))


class GattLibConnection(Connection):
    """
    Main transport class, uses real Bluetooth LE connection.
    Loops with timeout of 1 seconds to find device named "Lego MOVE Hub"

    :type requester: Requester
    """

    def __init__(self, bt_iface_name='hci0'):
        super().__init__()
        self.requester = None
        self._iface = bt_iface_name

    def connect(self, hub_mac=None, hub_name=None):
        service = DiscoveryService(self._iface)

        while not self.requester:
            log.info("Discovering devices using %s...", self._iface)
            devices = service.discover(1)
            log.debug("Devices: %s", devices)

            for address, name in devices.items():
                if self._is_device_matched(address, name, hub_mac, hub_name):
                    self.requester = Requester(address, True, self._iface)
                    break

            if self.requester:
                break

        return self

    def set_notify_handler(self, handler):
        if self.requester:
            log.debug("Setting notification handler: %s", handler)
            self.requester.notification_sink = handler
        else:
            raise RuntimeError("No requester available")

    def write(self, handle, data):
        log.debug("Writing to %s: %s", handle, str2hex(data))
        return self.requester.write_by_handle(handle, data)

    def is_alive(self):
        return self.requester.notify_thread.isAlive()



import logging
import re
import threading
from time import sleep

import gatt

from pylgbst.comms import Connection, MOVE_HUB_HW_UUID_SERV, MOVE_HUB_HW_UUID_CHAR, \
    MOVE_HUB_HARDWARE_HANDLE
from pylgbst.utilities import str2hex

log = logging.getLogger('comms-gatt')


class CustomDevice(gatt.Device):
    def __init__(self, mac_address, manager):
        gatt.Device.__init__(self, mac_address=mac_address, manager=manager)
        self._notify_callback = lambda hnd, val: None
        self._handle = None

    def connect(self):
        gatt.Device.connect(self)
        log.info("Waiting for device connection...")
        while self._handle is None:
            log.debug("Sleeping...")
            sleep(1)

        if isinstance(self._handle, BaseException):
            exc = self._handle
            self._handle = None
            raise exc

    def write(self, data):
        log.debug("Writing to handle: %s", str2hex(data))
        return self._handle.write_value(data)

    def enable_notifications(self):
        log.debug('Enable Notifications...')
        self._handle.enable_notifications()

    def set_notific_handler(self, func_hnd):
        self._notify_callback = func_hnd

    def services_resolved(self):
        log.debug('Getting MoveHub services and characteristics...')
        gatt.Device.services_resolved(self)
        log.debug("[%s] Resolved services", self.mac_address)
        for service in self.services:
            log.debug("[%s]  Service [%s]", self.mac_address, service.uuid)
            for characteristic in service.characteristics:
                log.debug("[%s]    Characteristic [%s]", self.mac_address, characteristic.uuid)
                if service.uuid == MOVE_HUB_HW_UUID_SERV and characteristic.uuid == MOVE_HUB_HW_UUID_CHAR:
                    log.debug('MoveHub characteristic found')
                    self._handle = characteristic

        if self._handle is None:
            self.manager.stop()
            self._handle = RuntimeError("Failed to obtain MoveHub handle")

    def characteristic_value_updated(self, characteristic, value):
        value = self._fix_weird_bug(value)
        log.debug('Notification in GattDevice: %s', str2hex(value))
        self._notify_callback(MOVE_HUB_HARDWARE_HANDLE, value)

    def _fix_weird_bug(self, value):
        if isinstance(value, str) and "dbus.Array" in value:  # weird bug from gatt on my Ubuntu 16.04!
            log.debug("Fixing broken notify string: %s", value)
            return ''.join([chr(int(x.group(1))) for x in re.finditer(r"dbus.Byte\((\d+)\)", value)])

        return value


class GattConnection(Connection):
    """
    :type _device: CustomDevice
    """

    def __init__(self, bt_iface_name='hci0'):
        super().__init__()
        self._device = None
        self._iface = bt_iface_name
        try:
            self._manager = gatt.DeviceManager(adapter_name=self._iface)
        except TypeError:
            raise NotImplementedError("Gatt is not implemented for this platform")

        self._manager_thread = threading.Thread(target=self._manager.run)
        self._manager_thread.setDaemon(True)
        log.debug('Starting DeviceManager...')

    def connect(self, hub_mac=None, hub_name=None):
        self._manager_thread.start()
        self._manager.start_discovery()

        while not self._device:
            log.info("Discovering devices...")
            devices = self._manager.devices()
            log.debug("Devices: %s", devices)

            for dev in devices:
                address = dev.mac_address
                name = dev.alias()
                if self._is_device_matched(address, name, hub_mac, hub_name):
                    self._device = CustomDevice(address, self._manager)
                    break

            if not self._device:
                sleep(1)

        self._device.connect()
        return self

    def disconnect(self):
        self._manager.stop()
        self._device.disconnect()

    def write(self, handle, data):
        self._device.write(data)

    def set_notify_handler(self, handler):
        self._device.set_notific_handler(handler)

    def enable_notifications(self):
        self._device.enable_notifications()

    def is_alive(self):
        return self._manager_thread.isAlive()



import threading

from pylgbst import get_connection_auto
from pylgbst.messages import *
from pylgbst.peripherals import *
from pylgbst.utilities import queue
from pylgbst.utilities import str2hex, usbyte, ushort

log = logging.getLogger("hub")

PERIPHERAL_TYPES = {
    DevTypes.MOTOR: Motor,
    DevTypes.SYSTEM_TRAIN_MOTOR: TrainMotor,
    DevTypes.MOTOR_EXTERNAL_TACHO: EncodedMotor,
    DevTypes.MOTOR_INTERNAL_TACHO: EncodedMotor,
    DevTypes.VISION_SENSOR: VisionSensor,
    DevTypes.RGB_LIGHT: LEDRGB,
    DevTypes.LED_LIGHT: LEDLight,
    DevTypes.TILT_EXTERNAL: TiltSensor,
    DevTypes.TILT_INTERNAL: TiltSensor,
    DevTypes.CURRENT: Current,
    DevTypes.VOLTAGE: Voltage,
    DevTypes.DUPLO_TRAIN_BASE_MOTOR.value: Motor,
    # DUPLO_TRAIN_BASE_SPEAKER = 42
    DevTypes.DUPLO_TRAIN_BASE_COLOR_SENSOR: VisionSensor,
    # DUPLO_TRAIN_BASE_SPEEDOMETER = 44
    DevTypes.TECHNIC_LARGE_LINEAR_MOTOR: EncodedMotor,
    DevTypes.TECHNIC_XLARGE_LINEAR_MOTOR: EncodedMotor,
    DevTypes.TECHNIC_MEDIUM_ANGULAR_MOTOR: EncodedMotor,
    DevTypes.TECHNIC_LARGE_ANGULAR_MOTOR: EncodedMotor,
    # DevTypes.TECHNIC_MEDIUM_HUB_GEST_SENSOR: 54
    DevTypes.REMOTE_CONTROL_BUTTON: RemoteButton,
    DevTypes.REMOTE_CONTROL_RSSI: Peripheral,
    # DevTypes.TECHNIC_MEDIUM_HUB_ACCELEROMETER: 57
    # DevTypes.TECHNIC_MEDIUM_HUB_GYRO_SENSOR: 58
    DevTypes.TECHNIC_MEDIUM_HUB_TILT_SENSOR: TiltSensor,
    DevTypes.TECHNIC_MEDIUM_HUB_TEMPERATURE_SENSOR: Temperature,
    DevTypes.TECHNIC_COLOR_SENSOR: VisionSensor,
    DevTypes.TECHNIC_DISTANCE_SENSOR: VisionSensor,
    # DevTypes.TECHNIC_FORCE_SENSOR: 63  # Spike Prime
    # DevTypes.MARIO_ACCELEROMETER: 71
    # DevTypes.MARIO_BARCODE_SENSOR: 73
    # DevTypes.MARIO_PANTS_SENSOR: 74
    DevTypes.TECHNIC_MEDIUM_ANGULAR_MOTOR_GREY: EncodedMotor,
    DevTypes.TECHNIC_LARGE_ANGULAR_MOTOR_GREY: EncodedMotor,
}


class Hub:
    """
    :type connection: pylgbst.comms.Connection
    :type peripherals: dict[int,Peripheral]
    """

    HUB_HARDWARE_HANDLE = 0x0E

    def __init__(self, connection=None):
        self._msg_handlers = []
        self.peripherals = {}
        self._sync_request = None
        self._sync_replies = queue.Queue(1)
        self._sync_lock = threading.Lock()

        self.add_message_handler(MsgHubAttachedIO, self._handle_device_change)
        self.add_message_handler(MsgPortOutputFeedback, self._handle_output_feedback)
        self.add_message_handler(MsgPortValueSingle, self._handle_sensor_data)
        self.add_message_handler(MsgPortValueCombined, self._handle_sensor_data)
        self.add_message_handler(MsgGenericError, self._handle_error)
        self.add_message_handler(MsgHubAction, self._handle_action)

        if not connection:
            connection = get_connection_auto()  # TODO: how to identify the hub?
        self.connection = connection
        self.connection.set_notify_handler(self._notify)
        self.connection.enable_notifications()

    def __del__(self):
        if self.connection and self.connection.is_alive():
            self.connection.disconnect()

    def add_message_handler(self, classname, callback):
        self._msg_handlers.append((classname, callback))

    def send(self, msg):
        """
        :type msg: pylgbst.messages.DownstreamMsg
        :rtype: pylgbst.messages.UpstreamMsg
        """
        log.debug("Send message: %r", msg)
        msgbytes = msg.bytes()
        if msg.needs_reply:
            with self._sync_lock:
                assert not self._sync_request, "Pending request %r while trying to put %r" % (self._sync_request, msg)
                self._sync_request = msg
                log.debug("Waiting for sync reply to %r...", msg)

            self.connection.write(self.HUB_HARDWARE_HANDLE, msgbytes)
            resp = self._sync_replies.get()
            log.debug("Fetched sync reply: %r", resp)
            if isinstance(resp, MsgGenericError):
                raise RuntimeError(resp.message())
            return resp
        else:
            self.connection.write(self.HUB_HARDWARE_HANDLE, msgbytes)
            return None

    def _notify(self, handle, data):
        log.debug("Notification on %s: %s", handle, str2hex(data))

        msg = self._get_upstream_msg(data)

        with self._sync_lock:
            if self._sync_request:
                if self._sync_request.is_reply(msg):
                    log.debug("Found matching upstream msg: %r", msg)
                    self._sync_replies.put(msg)
                    self._sync_request = None

        for msg_class, handler in self._msg_handlers:
            if isinstance(msg, msg_class):
                log.debug("Handling msg with %s: %r", handler, msg)
                handler(msg)

    def _get_upstream_msg(self, data):
        msg_type = usbyte(data, 2)
        msg = None
        for msg_kind in UPSTREAM_MSGS:
            if msg_type == msg_kind.TYPE:
                msg = msg_kind.decode(data)
                log.debug("Decoded message: %r", msg)
                break
        assert msg
        return msg

    def _handle_error(self, msg):
        log.warning("Command error: %s", msg.message())
        with self._sync_lock:
            if self._sync_request:
                self._sync_request = None
                self._sync_replies.put(msg)

    def _handle_action(self, msg):
        """
        :type msg: MsgHubAction
        """
        if msg.action == MsgHubAction.UPSTREAM_DISCONNECT:
            log.warning("Hub disconnects")
            self.connection.disconnect()
        elif msg.action == MsgHubAction.UPSTREAM_SHUTDOWN:
            log.warning("Hub switches off")
            self.connection.disconnect()

    def _handle_device_change(self, msg):
        if msg.event == MsgHubAttachedIO.EVENT_DETACHED:
            if msg.port not in self.peripherals:
                log.warning("Strange: got detach command for port %s that is not attached, will ignore it", msg.port)
            else:
                log.info("Detaching peripheral: %s", self.peripherals[msg.port])
                self.peripherals.pop(msg.port)
            return

        assert msg.event in (msg.EVENT_ATTACHED, msg.EVENT_ATTACHED_VIRTUAL)
        port = msg.port
        dev_type_raw = ushort(msg.payload, 0)
        dev_type = DevTypes(dev_type_raw) if DevTypes.has_value(dev_type_raw) else DevTypes.UNKNOWN

        if dev_type in PERIPHERAL_TYPES:
            self.peripherals[port] = PERIPHERAL_TYPES[dev_type](self, port)
        else:
            log.warning("Have no dedicated class for peripheral type 0x%x (%s) on port 0x%x",
                        dev_type_raw, DevTypes(dev_type).name, port)

            self.peripherals[port] = Peripheral(self, port)

        log.info("Attached peripheral %s => %s", DevTypes(dev_type).name, self.peripherals[msg.port])

        if msg.event == msg.EVENT_ATTACHED:
            hw_revision = reversed([usbyte(msg.payload, x) for x in range(2, 6)])
            sw_revision = reversed([usbyte(msg.payload, x) for x in range(6, 10)])
            # what to do with this info? it's useless, I guess
            del hw_revision, sw_revision
        elif msg.event == msg.EVENT_ATTACHED_VIRTUAL:
            self.peripherals[port].virtual_ports = (usbyte(msg.payload, 2), usbyte(msg.payload, 3))

    def _handle_output_feedback(self, msg):
        assert isinstance(msg, MsgPortOutputFeedback)
        if msg.port not in self.peripherals:
            log.warning("Notification on port with no device: %s", msg.port)
            return

        device = self.peripherals[msg.port]
        if hasattr(device, "cmd_in_progress"):
            device.cmd_in_progress = msg.is_in_progress()
            if device.cmd_in_progress:
                log.debug("Command on device %s in progress.", device)
            else:
                log.debug("Command on device %s completed.", device)

    def _handle_sensor_data(self, msg):
        assert isinstance(msg, (MsgPortValueSingle, MsgPortValueCombined))
        if msg.port not in self.peripherals:
            log.warning("Notification on port with no device: %s", msg.port)
            return

        device = self.peripherals[msg.port]
        device.queue_port_data(msg)

    def disconnect(self):
        self.send(MsgHubAction(MsgHubAction.DISCONNECT))

    def switch_off(self):
        self.send(MsgHubAction(MsgHubAction.SWITCH_OFF))


class MoveHub(Hub):
    """
    Class implementing Lego Boost's MoveHub specifics

    :type led: LEDRGB
    :type tilt_sensor: TiltSensor
    :type button: Button
    :type current: Current
    :type voltage: Voltage
    :type vision_sensor: pylgbst.peripherals.VisionSensor
    :type port_C: pylgbst.peripherals.Peripheral
    :type port_D: pylgbst.peripherals.Peripheral
    :type motor_A: EncodedMotor
    :type motor_B: EncodedMotor
    :type motor_AB: EncodedMotor
    :type motor_external: EncodedMotor
    """

    DEFAULT_NAME = "LEGO Move Hub"

    # PORTS
    PORT_A = 0x00
    PORT_B = 0x01
    PORT_C = 0x02
    PORT_D = 0x03
    PORT_AB = 0x10
    PORT_LED = 0x32
    PORT_TILT_SENSOR = 0x3A
    PORT_CURRENT = 0x3B
    PORT_VOLTAGE = 0x3C

    # noinspection PyTypeChecker
    def __init__(self, connection=None):
        self._comm_lock = threading.RLock()
        if connection is None:
            connection = get_connection_auto(hub_name=self.DEFAULT_NAME)

        super().__init__(connection)
        self.info = {}

        # shorthand fields
        self.button = Button(self)
        self.led = None
        self.current = None
        self.voltage = None
        self.motor_A = None
        self.motor_B = None
        self.motor_AB = None
        self.vision_sensor = None
        self.tilt_sensor = None
        self.motor_external = None
        self.port_C = None
        self.port_D = None

        self._wait_for_devices()
        self._report_status()

    def _wait_for_devices(self, get_dev_set=None):
        if not get_dev_set:
            get_dev_set = lambda: (self.motor_A, self.motor_B, self.motor_AB, self.led, self.tilt_sensor,
                                   self.current, self.voltage)
        for num in range(0, 100):
            devices = get_dev_set()
            if all(devices):
                log.debug("All devices are present: %s", devices)
                return
            log.debug("Waiting for builtin devices to appear: %s", devices)
            time.sleep(0.1)
        log.warning("Got only these devices: %s", get_dev_set())

    def _report_status(self):
        # maybe add firmware version
        name = self.send(MsgHubProperties(MsgHubProperties.ADVERTISE_NAME, MsgHubProperties.UPD_REQUEST))
        mac = self.send(MsgHubProperties(MsgHubProperties.PRIMARY_MAC, MsgHubProperties.UPD_REQUEST))
        log.info("%s on %s", name.payload, str2hex(mac.payload))

        voltage = self.send(MsgHubProperties(MsgHubProperties.VOLTAGE_PERC, MsgHubProperties.UPD_REQUEST))
        assert isinstance(voltage, MsgHubProperties)
        log.info("Voltage: %s%%", usbyte(voltage.parameters, 0))

        voltage = self.send(MsgHubAlert(MsgHubAlert.LOW_VOLTAGE, MsgHubAlert.UPD_REQUEST))
        assert isinstance(voltage, MsgHubAlert)
        if not voltage.is_ok():
            log.warning("Low voltage, check power source (maybe replace battery)")

    # noinspection PyTypeChecker
    def _handle_device_change(self, msg):
        with self._comm_lock:
            super()._handle_device_change(msg)
            if (
                    isinstance(msg, MsgHubAttachedIO)
                    and msg.event != MsgHubAttachedIO.EVENT_DETACHED
            ):
                port = msg.port
                if port == self.PORT_A:
                    self.motor_A = self.peripherals[port]
                elif port == self.PORT_B:
                    self.motor_B = self.peripherals[port]
                elif port == self.PORT_AB:
                    self.motor_AB = self.peripherals[port]
                elif port == self.PORT_C:
                    self.port_C = self.peripherals[port]
                elif port == self.PORT_D:
                    self.port_D = self.peripherals[port]
                elif port == self.PORT_LED:
                    self.led = self.peripherals[port]
                elif port == self.PORT_TILT_SENSOR:
                    self.tilt_sensor = self.peripherals[port]
                elif port == self.PORT_CURRENT:
                    self.current = self.peripherals[port]
                elif port == self.PORT_VOLTAGE:
                    self.voltage = self.peripherals[port]

                if type(self.peripherals[port]) == VisionSensor:
                    self.vision_sensor = self.peripherals[port]
                elif type(self.peripherals[port]) == EncodedMotor and port not in (
                        self.PORT_A,
                        self.PORT_B,
                        self.PORT_AB,
                ):
                    self.motor_external = self.peripherals[port]


class SmartHub(Hub):
    """
    Class implementing Lego SmartHub specifics
    https://www.lego.com/en-pt/product/hub-88009

    :type led: LEDRGB
    :type current: Current
    :type voltage: Voltage
    :type port_A: Peripheral
    :type port_B: Peripheral
    """

    DEFAULT_NAME = "Smart Hub"

    # PORTS
    PORT_A = 0x00
    PORT_B = 0x01
    PORT_LED = 0x32
    PORT_CURRENT = 0x3B
    PORT_VOLTAGE = 0x3C

    def __init__(self, connection=None):
        if connection is None:
            connection = get_connection_auto(hub_name=self.DEFAULT_NAME)

        super().__init__(connection)

        self.button = Button(self)
        self.led = None
        self.port_A = None
        self.port_B = None
        self.current = None
        self.voltage = None

        self._wait_for_devices()

    def _wait_for_devices(self, get_dev_set=None):
        if not get_dev_set:
            get_dev_set = lambda: (self.led, self.current, self.voltage)

        for num in range(0, 100):
            devices = get_dev_set()
            if all(devices):
                log.debug("All devices are present: %s", devices)
                return
            log.debug("Waiting for builtin devices to appear: %s", devices)
            time.sleep(0.1)
        log.warning("Got only these devices: %s", get_dev_set())

    # noinspection PyTypeChecker
    def _handle_device_change(self, msg):
        super()._handle_device_change(msg)
        if (
                isinstance(msg, MsgHubAttachedIO)
                and msg.event != MsgHubAttachedIO.EVENT_DETACHED
        ):
            port = msg.port
            if port == self.PORT_A:
                self.port_A = self.peripherals[port]
            elif port == self.PORT_B:
                self.port_B = self.peripherals[port]
            elif port == self.PORT_LED:
                self.led = self.peripherals[port]
            elif port == self.PORT_CURRENT:
                self.current = self.peripherals[port]
            elif port == self.PORT_VOLTAGE:
                self.voltage = self.peripherals[port]


class RemoteHandset(Hub):
    """
    Class implementing remote handset specifics

    :type led: LEDRGB
    :type current: Current
    :type voltage: Voltage
    :type port_A: RemoteButton
    :type port_B: RemoteButton
    """

    DEFAULT_NAME = "Remote Handset"

    # PORTS
    PORT_A = 0x00
    PORT_B = 0x01
    PORT_LED = 0x34
    PORT_VOLTAGE = 0x3B
    PORT_RSSI = 0x3C

    def __init__(self, connection=None, address=None):
        if connection is None:
            connection = get_connection_auto(hub_mac=address, hub_name=self.DEFAULT_NAME)

        super().__init__(connection)

        self.led = None
        self.port_A = None
        self.port_B = None
        self.port_RSSI = None
        self.voltage = None

        self._wait_for_devices()

    def _wait_for_devices(self, get_dev_set=None):
        if not get_dev_set:
            get_dev_set = lambda: (self.port_A, self.port_B, self.port_RSSI, self.led, self.voltage)

        for num in range(0, 100):
            devices = get_dev_set()
            if all(devices):
                log.debug("All devices are present: %s", devices)
                return
            log.debug("Waiting for builtin devices to appear: %s", devices)
            time.sleep(0.1)
        log.warning("Got only these devices: %s", get_dev_set())

    # noinspection PyTypeChecker
    def _handle_device_change(self, msg):
        super()._handle_device_change(msg)
        if (
                isinstance(msg, MsgHubAttachedIO)
                and msg.event != MsgHubAttachedIO.EVENT_DETACHED
        ):
            port = msg.port
            if port == self.PORT_A:
                self.port_A = self.peripherals[port]
            elif port == self.PORT_B:
                self.port_B = self.peripherals[port]
            elif port == self.PORT_LED:
                self.led = self.peripherals[port]
            elif port == self.PORT_VOLTAGE:
                self.voltage = self.peripherals[port]
            elif port == self.PORT_RSSI:
                self.port_RSSI = self.peripherals[port]



import logging
import traceback

from pylgbst.comms import DebugServer, Connection

log = logging.getLogger('pylgbst')


def get_connection_bluegiga(controller=None, hub_mac=None, hub_name=None):
    del controller  # to prevent code analysis warning
    from pylgbst.comms.cpygatt import BlueGigaConnection

    return BlueGigaConnection().connect(hub_mac, hub_name)


def get_connection_gattool(controller='hci0', hub_mac=None, hub_name=None):
    from pylgbst.comms.cpygatt import GattoolConnection

    return GattoolConnection(controller).connect(hub_mac, hub_name)


def get_connection_gatt(controller='hci0', hub_mac=None, hub_name=None):
    from pylgbst.comms.cgatt import GattConnection

    return GattConnection(controller).connect(hub_mac, hub_name)


def get_connection_gattlib(controller='hci0', hub_mac=None, hub_name=None):
    from pylgbst.comms.cgattlib import GattLibConnection

    return GattLibConnection(controller).connect(hub_mac, hub_name)


def get_connection_bluepy(controller='hci0', hub_mac=None, hub_name=None):
    from pylgbst.comms.cbluepy import BluepyConnection

    return BluepyConnection(controller).connect(hub_mac, hub_name)


def get_connection_bleak(controller='hci0', hub_mac=None, hub_name=None):
    """
    Return connection based with Bleak API as an endpoint.

    :param controller: Not used, kept for compatibility with others.
    :param hub_mac: Optional Lego HUB MAC to connect to.
    :return: Driver object.
    """
    del controller  # to prevent code analysis warning
    from pylgbst.comms.cbleak import BleakDriver

    return BleakDriver(hub_mac, hub_name)


def get_connection_auto(controller='hci0', hub_mac=None, hub_name=None) -> Connection:
    fns = [
        get_connection_bleak,
        get_connection_bluepy,
        get_connection_bluegiga,
        get_connection_gatt,
        get_connection_gattool,
        get_connection_gattlib,
    ]

    conn = None
    for fn in fns:
        try:
            logging.info("Trying %s", fn.__name__)
            return fn(controller, hub_mac, hub_name)
        except KeyboardInterrupt:
            raise
        except BaseException:
            logging.debug("Failed: %s", traceback.format_exc())

    if conn is None:
        raise Exception("Failed to autodetect connection, make sure you have installed prerequisites")

    logging.info("Succeeded with %s", conn.__class__.__name__)
    return conn


def start_debug_server(iface="hci0", port=9090):
    server = DebugServer(get_connection_auto(iface))
    try:
        server.start(port)
    finally:
        server.connection.disconnect()



import asyncio
import logging
import platform
import queue
import threading
import time

import bleak

from pylgbst.comms import Connection, MOVE_HUB_HW_UUID_CHAR, MOVE_HUB_HW_UUID_SERV

log = logging.getLogger('comms-bleak')


class BleakDriver(Connection):
    """Driver that provides interface between API and Bleak."""

    def __init__(self, hub_mac=None, hub_name=None):
        """
        Initialize new object of Bleak Driver class.

        :param hub_mac: Optional Lego HUB MAC to connect to.
        """
        self.hub_mac = hub_mac
        self.hub_name = hub_name
        self._handler = None
        self._abort = False
        self._connection_thread = None
        self._processing_thread = None

        # Queues to handle request / responses. Acts as a buffer between API and async BLE driver
        self.resp_queue = queue.Queue()
        self.req_queue = queue.Queue()

    def set_notify_handler(self, handler):
        """
        Set handler function used to communicate with an API.

        :param handler: Handler function called by driver when received data
        :return: None
        """
        self._handler = handler

    def enable_notifications(self):
        """
        Enable notifications, in our cases starts communication threads.

        We cannot do this earlier, because API need to fist set notification handler.
        :return: None
        """
        self._connection_thread = threading.Thread(target=lambda: asyncio.run(self._bleak_thread()))
        self._connection_thread.daemon = True
        self._connection_thread.start()

        self._processing_thread = threading.Thread(target=self._processing)
        self._processing_thread.daemon = True
        self._processing_thread.start()

    async def _bleak_thread(self):
        bleak = BleakConnection()
        # For MacOS 12+ the service_uuids kwarg is required for scanning
        kwargs = {}
        if "Darwin" == platform.system() and int(platform.mac_ver()[0].split(".")[0]) >= 12:
            kwargs = {"service_uuids": [MOVE_HUB_HW_UUID_SERV]}
        await bleak.connect(self.hub_mac, self.hub_name, **kwargs)
        await bleak.set_notify_handler((self._safe_handler, self.resp_queue))
        # After connecting, need to send any data or hub will drop the connection,
        # below command is Advertising name request update
        await bleak.write_char(MOVE_HUB_HW_UUID_CHAR, bytearray([0x05, 0x00, 0x01, 0x01, 0x05]))
        while not self._abort:
            await asyncio.sleep(0.1)
            if self.req_queue.qsize() != 0:
                data = self.req_queue.get()
                await bleak.write(data[0], data[1])

        logging.info("Communications thread has exited")

    @staticmethod
    def _safe_handler(handler, data, resp_queue):
        resp_queue.put((handler, data))

    def _processing(self):
        while not self._abort:
            if self.resp_queue.qsize() != 0:
                msg = self.resp_queue.get()
                self._handler(msg[0], bytes(msg[1]))

            time.sleep(0.01)
        logging.info("Processing thread has exited")

    def write(self, handle, data):
        """
        Send data to given handle number.

        :param handle: Handle number that will be translated into characteristic uuid
        :param data: data to send
        :raises ConnectionError" When internal threads are not working
        :return: None
        """
        if not self._connection_thread.is_alive() or not self._processing_thread.is_alive():
            raise ConnectionError('Something went wrong, communication threads not functioning.')

        self.req_queue.put((handle, data))

    def disconnect(self):
        """
        Disconnect and stops communication threads.

        :return: None
        """
        self._abort = True

    def is_alive(self):
        """
        Indicate whether driver is functioning or not.

        :return: True if driver is functioning; False otherwise.
        """
        if self._connection_thread is not None and self._processing_thread is not None:
            return self._connection_thread.is_alive() and self._processing_thread.is_alive()
        else:
            return False

class BleakConnection(Connection):
    """Bleak driver for communicating with BLE device."""

    def __init__(self):
        """Initialize new instance of BleakConnection class."""
        Connection.__init__(self)

        self._device = None
        self._client = None
        logging.getLogger('bleak.backends.dotnet.client').setLevel(logging.WARNING)
        logging.getLogger('bleak.backends.bluezdbus.client').setLevel(logging.WARNING)

    async def find_device(self, hub_mac=None, hub_name=None):
        found_event = asyncio.Event()
        found_device = None

        async def _detect(device: bleak.BLEDevice, advertisement_data: bleak.AdvertisementData):
            nonlocal found_device
            nonlocal found_event
            address = device.address
            name = device.name
            if self._is_device_matched(address, name, hub_mac, hub_name):
                log.info('Device matched: %r', device)
                found_device = device
                found_event.set()
        
        scanner = bleak.BleakScanner(
            detection_callback=_detect
        )
        await scanner.start()
        await found_event.wait()
        await scanner.stop()
        return found_device

    async def connect(self, hub_mac=None, hub_name=None, **kwargs):
        """
        Connect to device.

        :param hub_mac: Optional Lego HUB MAC to connect to
        :param hub_name: Optional Lego Hub name to connect to
        :kwargs: Optional parameters for bleak.discover

        :raises ConnectionError: When cannot connect to given MAC or name matching fails.
        :return: None
        """
        log.info("Discovering devices... Press green button on Hub")

        self._device = await self.find_device(hub_mac, hub_name, **kwargs)

        self._client = bleak.BleakClient(self._device)
        status = await self._client.connect()
        log.debug('Connection status: {status}'.format(status=status))

    async def write(self, handle, data):
        """
        Send data to given handle number.

        If handle cannot be found in service description, hardcoded LEGO uuid will be used.
        :param handle: Handle number that will be translated into characteristic uuid
        :param data: data to send
        :return: None
        """
        log.debug('Request: {handle} {payload}'.format(handle=handle, payload=[hex(x) for x in data]))
        desc = self._client.services.get_descriptor(handle)

        if not isinstance(data, bytearray):
            data = bytearray(data)

        if desc is None:
            # dedicated handle not found, try to send by using LEGO Move Hub default characteristic
            await self._client.write_gatt_char(MOVE_HUB_HW_UUID_CHAR, data)
        else:
            await self._client.write_gatt_char(desc.characteristic_uuid, data)

    async def write_char(self, characteristic_uuid, data):
        """
        Send data to given handle number.

        :param characteristic_uuid: Characteristic uuid used to send data
        :param data: data to send
        :return: None
        """
        await self._client.write_gatt_char(characteristic_uuid, data)

    async def set_notify_handler(self, inputs):
        """
        Set notification handler.

        :param handler: Handle function to be called when receive any data.
        :return: None
        """
        handler, resp_queue = inputs

        def c(handle, data):
            log.debug('Response: {handle} {payload}'.format(handle=handle, payload=[hex(x) for x in data]))
            handler(handle, data, resp_queue)

        await self._client.start_notify(MOVE_HUB_HW_UUID_CHAR, c)

    def is_alive(self):
        """
        To keep compatibility with the driver interface.

        This method does nothing.
        :return: None.
        """
        pass



import unittest

from pylgbst.comms import *


class ConnectionTestCase(unittest.TestCase):
    def test_is_device_matched(self):
        conn = Connection()

        hub_address = '1a:2A:3A:4A:5A:6A'
        other_address = 'A1:a2:a3:a4:a5:a6'
        zero_address = '00:00:00:00:00:00'
        hub_name = 'LEGO Move Hub'
        other_name = 'HRM'

        test_matrix = [
            # address,      name,       hub_mac,        expected
            (hub_address, hub_name, hub_address, None, True),
            (hub_address, hub_name, None, hub_name, True),
            (hub_address, None, hub_address, None, True),
            (hub_address, None, None, hub_name, False),
            (hub_address, other_name, hub_address, None, True),
            (hub_address, other_name, None, hub_name, False),
            (other_address, hub_name, hub_address, None, False),
            (other_address, hub_name, None, hub_name, True),
            (other_address, None, hub_address, None, False),
            (other_address, None, None, hub_name, False),
            (other_address, other_name, hub_address, None, False),
            (other_address, other_name, None, hub_name, False),
            (zero_address, hub_name, hub_address, None, False),
            (zero_address, hub_name, None, hub_name, False),
            (zero_address, None, hub_address, None, False),
            (zero_address, None, None, hub_name, False),
            (zero_address, other_name, hub_address, None, False),
            (zero_address, other_name, None, hub_name, False),
        ]

        for address, name, hub_mac, fname, expected in test_matrix:
            matched = conn._is_device_matched(address=address, dev_name=name, hub_mac=hub_mac, find_name=fname)
            self.assertEqual(matched, expected)



import dbus
import sys
import unittest

from gatt import DeviceManager

from pylgbst.comms.cgatt import CustomDevice, GattConnection
from tests import log, str2hex


class MockBus(object):
    def __init__(self, *args, **kwargs):
        # super(MockBus, self).__init__(*args, **kwargs)
        pass

    def get_object(self, bus_name, object_path, introspect=True, follow_name_owner_changes=False, **kwargs):
        return None


dbus.SystemBus = lambda: MockBus()


class DeviceManagerMock(DeviceManager, object):
    def update_devices(self):
        pass


class TestGatt(unittest.TestCase):
    def test_one(self):
        log.debug("")
        manager = DeviceManagerMock("hci0")
        obj = CustomDevice("AA", manager)

        def callback(handle, value):
            log.debug("%s: %s", type(value), str2hex(value))
            if sys.version_info[0] == 2:
                self.assertEquals("0f0004020126000000001000000010", str2hex(value))

        obj.set_notific_handler(callback)
        arr = "dbus.Array([dbus.Byte(15), dbus.Byte(0), dbus.Byte(4), dbus.Byte(2), dbus.Byte(1), dbus.Byte(38), " \
              "dbus.Byte(0), dbus.Byte(0), dbus.Byte(0), dbus.Byte(0), dbus.Byte(16), dbus.Byte(0), dbus.Byte(0), " \
              "dbus.Byte(0), dbus.Byte(16)], signature=dbus.Signature('y'), variant_level=1)"
        obj.characteristic_value_updated(None, arr if sys.version_info[0] == 2 else bytes(arr, 'ascii'))

    def test_conn(self):
        try:
            obj = GattConnection()
            obj.connect()
        except AttributeError:
            pass



import logging
from enum import Enum, unique
from struct import pack, unpack

from pylgbst.utilities import str2hex

log = logging.getLogger("hub")


class Message:
    TYPE = None

    def __init__(self):
        self.hub_id = 0x00  # not used according to official doc
        self.payload = b""

    def bytes(self):
        """
        see https://lego.github.io/lego-ble-wireless-protocol-docs/#common-message-header
        """
        msglen = len(self.payload) + 3
        assert msglen < 127, "TODO: handle longer messages with 2-byte len"
        return pack("<B", msglen) + pack("<B", self.hub_id) + pack("<B", self.TYPE) + self.payload

    def __repr__(self):
        # assert self.bytes()  # to trigger any field changes
        data = self.__dict__
        data = {
            x: (str2hex(y) if isinstance(y, bytes) else y)
            for x, y in data.items()
            if x not in ("hub_id",)
        }
        return self.__class__.__name__ + "(%s)" % data


class DownstreamMsg(Message):
    def __init__(self):
        super().__init__()
        self.needs_reply = False

    def is_reply(self, msg):
        del msg
        return False


class UpstreamMsg(Message):
    def __init__(self):
        super().__init__()

    @classmethod
    def decode(cls, data):
        """
        see https://lego.github.io/lego-ble-wireless-protocol-docs/#common-message-header
        """
        msg = cls()
        msg.payload = data
        msglen = msg._byte()
        assert msglen < 127, "TODO: handle longer messages with 2-byte len"
        hub_id = msg._byte()
        assert hub_id == 0
        msg_type = msg._byte()
        assert cls.TYPE == msg_type, "Message type does not match: %x!=%x" % (cls.TYPE, msg_type)
        assert isinstance(msg.payload, (bytes, bytearray))
        return msg

    def __shift(self, vtype, vlen):
        val = self.payload[0:vlen]
        self.payload = self.payload[vlen:]
        return unpack("<" + vtype, val)[0]

    def _byte(self):
        return self.__shift("B", 1)

    def _short(self):
        return self.__shift("H", 2)

    def _long(self):
        return self.__shift("I", 4)

    def _float(self):
        return self.__shift("f", 4)

    def _bits_list(self, val):
        res = []
        x = 1
        for i in range(16 + 1):
            if val & x:
                res.append(i)
            x <<= 1
        return res


class MsgHubProperties(DownstreamMsg, UpstreamMsg):
    """
    https://lego.github.io/lego-ble-wireless-protocol-docs/index.html#hub-properties
    """

    TYPE = 0x01

    ADVERTISE_NAME = 0x01
    BUTTON = 0x02
    FW_VERSION = 0x03
    HW_VERSION = 0x04
    RSSI = 0x05
    VOLTAGE_PERC = 0x06
    BATTERY_TYPE = 0x07
    MANUFACTURER = 0x08
    RADIO_FW_VERSION = 0x09
    WIRELESS_PROTO_VERSION = 0x0A
    SYSTEM_TYPE_ID = 0x0B
    HW_NETW_ID = 0x0C
    PRIMARY_MAC = 0x0D
    SECONDARY_MAC = 0x0E
    HARDWARE_NETWORK_FAMILY = 0x0F

    SET = 0x01
    UPD_ENABLE = 0x02
    UPD_DISABLE = 0x03
    RESET = 0x04
    UPD_REQUEST = 0x05
    UPSTREAM_UPDATE = 0x06

    def __init__(self, prop=None, operation=None, parameters=b""):
        super().__init__()

        self.property = prop
        self.operation = operation
        self.parameters = parameters

    def bytes(self):
        if self.operation in (self.UPD_REQUEST, self.UPD_ENABLE):
            self.needs_reply = True
        self.payload = pack("<B", self.property) + pack("<B", self.operation) + self.parameters
        return super().bytes()

    @classmethod
    def decode(cls, data):
        msg = super().decode(data)
        assert isinstance(msg, MsgHubProperties)
        msg.property = msg._byte()
        msg.operation = msg._byte()
        msg.parameters = msg.payload
        return msg

    def is_reply(self, msg):
        return (
            isinstance(msg, MsgHubProperties)
            and msg.operation == self.UPSTREAM_UPDATE
            and msg.property == self.property
        )


class MsgHubAction(DownstreamMsg, UpstreamMsg):
    """
    https://lego.github.io/lego-ble-wireless-protocol-docs/index.html#hub-actions
    """

    TYPE = 0x02

    SWITCH_OFF = 0x01
    DISCONNECT = 0x02
    VCC_PORT_CONTROL_ON = 0x03
    VCC_PORT_CONTROL_OFF = 0x04
    BUSY_INDICATION_ON = 0x05
    BUSY_INDICATION_OFF = 0x06
    SWITCH_OFF_IMMEDIATELY = 0x2F

    UPSTREAM_SHUTDOWN = 0x30
    UPSTREAM_DISCONNECT = 0x31
    UPSTREAM_BOOT_MODE = 0x32

    def __init__(self, action=None):
        super().__init__()
        self.action = action

    def bytes(self):
        self.payload = pack("<B", self.action)
        self.needs_reply = self.action in (self.DISCONNECT, self.SWITCH_OFF)
        return super().bytes()

    def is_reply(self, msg):
        if not isinstance(msg, MsgHubAction):
            raise TypeError("Unexpected message type: %s" % (msg.__class__,))
        if self.action == self.DISCONNECT and msg.action == self.UPSTREAM_DISCONNECT:
            return True

        if self.action == self.SWITCH_OFF and msg.action == self.UPSTREAM_SHUTDOWN:
            return True

    @classmethod
    def decode(cls, data):
        msg = super().decode(data)
        assert isinstance(msg, MsgHubAction)
        msg.action = msg._byte()
        return msg


class MsgHubAlert(DownstreamMsg, UpstreamMsg):
    """
    https://lego.github.io/lego-ble-wireless-protocol-docs/index.html#hub-alerts
    """

    TYPE = 0x03

    LOW_VOLTAGE = 0x01
    HIGH_CURRENT = 0x02
    LOW_SIGNAL = 0x03
    OVER_POWER = 0x04

    DESCR = {
        LOW_VOLTAGE: "low voltage",
        HIGH_CURRENT: "high current",
        LOW_SIGNAL: "low signal",
        OVER_POWER: "over power",
    }

    UPD_ENABLE = 0x01
    UPD_DISABLE = 0x02
    UPD_REQUEST = 0x03
    UPSTREAM_UPDATE = 0x04

    def __init__(self, atype=None, operation=None):
        super().__init__()
        self.atype = atype
        self.operation = operation
        self.status = None

    def bytes(self):
        self.payload = pack("<B", self.atype) + pack("<B", self.operation)
        if self.operation == self.UPD_REQUEST:
            self.needs_reply = True
        return super().bytes()

    @classmethod
    def decode(cls, data):
        msg = super().decode(data)
        assert isinstance(msg, MsgHubAlert)
        msg.atype = msg._byte()
        msg.operation = msg._byte()
        msg.status = msg._byte()

        assert msg.operation == cls.UPSTREAM_UPDATE
        return msg

    def is_ok(self):
        return not self.status

    def is_reply(self, msg):
        return (
            isinstance(msg, MsgHubAlert)
            and msg.operation == self.UPSTREAM_UPDATE
            and msg.atype == self.atype
        )


@unique
class DevTypes(Enum):
    # thankfully borrowed some knowledge
    # from https://github.com/nathankellenicki/node-poweredup/blob/master/src/consts.ts

    UNKNOWN = 0x0000

    # TODO: either dec or hex, not both
    MOTOR = 0x0001
    SYSTEM_TRAIN_MOTOR = 0x0002
    BUTTON = 0x0005
    LED_LIGHT = 0x0008
    VOLTAGE = 0x0014
    CURRENT = 0x0015
    PIEZO_SOUND = 0x0016
    RGB_LIGHT = 0x0017
    TILT_EXTERNAL = 0x0022
    MOTION_SENSOR = 0x0023
    VISION_SENSOR = 0x0025
    MOTOR_EXTERNAL_TACHO = 0x0026
    MOTOR_INTERNAL_TACHO = 0x0027
    TILT_INTERNAL = 0x0028

    DUPLO_TRAIN_BASE_MOTOR = 41
    DUPLO_TRAIN_BASE_SPEAKER = 42
    DUPLO_TRAIN_BASE_COLOR_SENSOR = 43
    DUPLO_TRAIN_BASE_SPEEDOMETER = 44
    TECHNIC_LARGE_LINEAR_MOTOR = 46  # Technic Control+
    TECHNIC_XLARGE_LINEAR_MOTOR = 47  # Technic Control+
    TECHNIC_MEDIUM_ANGULAR_MOTOR = 48  # Spike Prime
    TECHNIC_LARGE_ANGULAR_MOTOR = 49  # Spike Prime
    TECHNIC_MEDIUM_HUB_GEST_SENSOR = 54
    REMOTE_CONTROL_BUTTON = 55
    REMOTE_CONTROL_RSSI = 56
    TECHNIC_MEDIUM_HUB_ACCELEROMETER = 57
    TECHNIC_MEDIUM_HUB_GYRO_SENSOR = 58
    TECHNIC_MEDIUM_HUB_TILT_SENSOR = 59
    TECHNIC_MEDIUM_HUB_TEMPERATURE_SENSOR = 60
    TECHNIC_COLOR_SENSOR = 61  # Spike Prime
    TECHNIC_DISTANCE_SENSOR = 62  # Spike Prime
    TECHNIC_FORCE_SENSOR = 63  # Spike Prime
    MARIO_ACCELEROMETER = 71
    MARIO_BARCODE_SENSOR = 73
    MARIO_PANTS_SENSOR = 74
    TECHNIC_MEDIUM_ANGULAR_MOTOR_GREY = 75  # Mindstorms
    TECHNIC_LARGE_ANGULAR_MOTOR_GREY = 76  # Technic Control+

    @classmethod
    def has_value(cls, value):
        values = set(item.value for item in cls.__members__.values())
        return value in values


class MsgHubAttachedIO(UpstreamMsg):
    """
    https://lego.github.io/lego-ble-wireless-protocol-docs/index.html#hub-attached-i-o
    """

    TYPE = 0x04

    EVENT_DETACHED = 0x00
    EVENT_ATTACHED = 0x01
    EVENT_ATTACHED_VIRTUAL = 0x02

    def __init__(self):
        super().__init__()
        self.port = None
        self.event = None

    @classmethod
    def decode(cls, data):
        msg = super().decode(data)
        assert isinstance(msg, MsgHubAttachedIO)
        msg.port = msg._byte()
        msg.event = msg._byte()
        return msg


class MsgGenericError(UpstreamMsg):
    """
    https://lego.github.io/lego-ble-wireless-protocol-docs/index.html#generic-error-messages
    """

    TYPE = 0x05

    ERR_ACK = 0x01  # ACK
    ERR_MACK = 0x02  # MACK
    ERR_BUFFER_OVERFLOW = 0x03  # Buffer Overflow
    ERR_TIMEOUT = 0x04  # Timeout
    ERR_WRONG_COMMAND = 0x05  # Command NOT recognized
    ERR_WRONG_PARAMS = 0x06  # Invalid use (e.g. parameter error(s)
    ERR_OVERCURRENT = 0x07
    ERR_INTERNAL = 0x08

    DESCR = {
        ERR_ACK: "ACK",
        ERR_MACK: "MACK",
        ERR_BUFFER_OVERFLOW: "Buffer Overflow",
        ERR_TIMEOUT: "Timeout",
        ERR_WRONG_COMMAND: "Command NOT recognized",
        ERR_WRONG_PARAMS: "Invalid use (e.g. parameter error(s)",
        ERR_OVERCURRENT: "Overcurrent",
        ERR_INTERNAL: "Internal ERROR",
    }

    def __init__(self):
        super().__init__()
        self.cmd = None
        self.err = None

    @classmethod
    def decode(cls, data):
        msg = super().decode(data)
        assert isinstance(msg, MsgGenericError)
        msg.cmd = msg._byte()
        msg.err = msg._byte()
        return msg

    def message(self):
        return "Command 0x%x caused error 0x%x: %s" % (self.cmd, self.err, self.DESCR[self.err])


class MsgPortInfoRequest(DownstreamMsg):
    """
    This is sync request for value on port
    https://lego.github.io/lego-ble-wireless-protocol-docs/index.html#port-information-request
    """

    TYPE = 0x21

    INFO_PORT_VALUE = 0x00
    INFO_MODE_INFO = 0x01
    INFO_MODE_COMBINATIONS = 0x02

    def __init__(self, port, info_type):
        super().__init__()
        self.port = port
        self.info_type = info_type
        self.needs_reply = True

    def bytes(self):
        self.payload = pack("<B", self.port) + pack("<B", self.info_type)
        return super().bytes()

    def is_reply(self, msg):
        if msg.port != self.port:
            return False

        if self.info_type == self.INFO_PORT_VALUE:
            return isinstance(msg, (MsgPortValueSingle, MsgPortValueCombined))
        else:
            return isinstance(msg, (MsgPortInfo,))


class MsgPortModeInfoRequest(DownstreamMsg):
    """
    https://lego.github.io/lego-ble-wireless-protocol-docs/index.html#port-mode-information-request
    """

    TYPE = 0x22

    INFO_NAME = 0x00
    INFO_RAW_RANGE = 0x01
    INFO_PCT_RANGE = 0x02
    INFO_SI_RANGE = 0x03  # no idea what 'SI' stands for
    INFO_UNITS = 0x04
    INFO_MAPPING = 0x05
    # INFO_INTERNAL = 0x06
    INFO_MOTOR_BIAS = 0x07
    INFO_CAPABILITY_BITS = 0x08
    INFO_VALUE_FORMAT = 0x80

    INFO_TYPES = {
        INFO_NAME: "Name",
        INFO_RAW_RANGE: "Raw range",
        INFO_PCT_RANGE: "Percent range",
        INFO_SI_RANGE: "SI value range",
        INFO_UNITS: "Units",
        INFO_MAPPING: "Mapping",
        INFO_MOTOR_BIAS: "Motor bias",
        INFO_CAPABILITY_BITS: "Capabilities",
        INFO_VALUE_FORMAT: "Value encoding",
    }

    def __init__(self, port, mode, info_type):
        super().__init__()
        self.port = port
        self.mode = mode
        self.info_type = info_type
        self.payload = pack("<B", port) + pack("<B", mode) + pack("<B", info_type)
        self.needs_reply = True

    def is_reply(self, msg):
        if not isinstance(msg, MsgPortModeInfo):
            return False

        if (
            msg.port != self.port
            or msg.mode != self.mode
            or msg.info_type != self.info_type
        ):
            return False

        return True


class MsgPortInputFmtSetupSingle(DownstreamMsg):
    """
    https://lego.github.io/lego-ble-wireless-protocol-docs/index.html#port-input-format-setup-single
    """

    TYPE = 0x41

    def __init__(self, port, mode, delta=1, update_enable=0):
        super().__init__()
        self.port = port
        self.mode = mode
        self.updates_enabled = update_enable
        self.update_delta = delta
        self.payload = pack("<B", port) + pack("<B", mode) + pack("<I", delta) + pack("<B", update_enable)
        self.needs_reply = True

    def is_reply(self, msg):
        if isinstance(msg, MsgPortInputFmtSingle) and msg.port == self.port:
            return True


class MsgPortInputFmtSetupCombined(DownstreamMsg):
    """
    https://lego.github.io/lego-ble-wireless-protocol-docs/index.html#port-input-format-setup-combinedmode
    """
    TYPE = 0x42

    def __init__(self, port, mode, delta=1, update_enable=0):
        super().__init__()
        self.port = port
        self.payload = pack("<B", port) + pack("<B", mode) + pack("<I", delta) + pack("<B", update_enable)
        self.needs_reply = True

    def is_reply(self, msg):
        if isinstance(msg, MsgPortInputFmtCombined) and msg.port == self.port:
            return True


class MsgPortInfo(UpstreamMsg):
    """
    https://lego.github.io/lego-ble-wireless-protocol-docs/index.html#port-information
    """

    TYPE = 0x43

    CAP_OUTPUT = 0b00000001
    CAP_INPUT = 0b00000010
    CAP_COMBINABLE = 0b00000100
    CAP_SYNCHRONIZABLE = 0b00001000

    def __init__(self):
        super().__init__()
        self.port = None
        self.info_type = None
        self.capabilities = None
        self.total_modes = None
        self.input_modes = None
        self.output_modes = None
        self.possible_mode_combinations = []

    @classmethod
    def decode(cls, data):
        msg = super().decode(data)
        assert isinstance(msg, MsgPortInfo)
        msg.port = msg._byte()
        msg.info_type = msg._byte()
        if msg.info_type == MsgPortInfoRequest.INFO_MODE_INFO:
            msg.capabilities = msg._byte()
            msg.total_modes = msg._byte()
            msg.input_modes = msg._bits_list(msg._short())
            msg.output_modes = msg._bits_list(msg._short())
        else:
            while msg.payload:
                # https://lego.github.io/lego-ble-wireless-protocol-docs/index.html#pos-m
                val = msg._short()
                msg.possible_mode_combinations.append(msg._bits_list(val))
                if not val:
                    break
        return msg

    def is_output(self):
        assert self.info_type == MsgPortInfoRequest.INFO_MODE_INFO
        return bool(self.capabilities & self.CAP_OUTPUT)

    def is_input(self):
        assert self.info_type == MsgPortInfoRequest.INFO_MODE_INFO
        return bool(self.capabilities & self.CAP_INPUT)

    def is_combinable(self):
        assert self.info_type == MsgPortInfoRequest.INFO_MODE_INFO
        return bool(self.capabilities & self.CAP_COMBINABLE)

    def is_synchronizable(self):
        assert self.info_type == MsgPortInfoRequest.INFO_MODE_INFO
        return bool(self.capabilities & self.CAP_SYNCHRONIZABLE)


class MsgPortModeInfo(UpstreamMsg):
    """
    https://lego.github.io/lego-ble-wireless-protocol-docs/index.html#port-mode-information
    """

    TYPE = 0x44

    MAPPING_FLAGS = {
        7: "Supports NULL value",
        6: "Supports Functional Mapping 2.0+",
        5: "N/A",
        4: "Absolute [min..max]",
        3: "Relative [-1..1]",
        2: "Discrete [0, 1, 2, 3]",
        1: "N/A",
        0: "N/A",
    }

    DATASET_TYPES = {
        0b00: "8 bit",
        0b01: "16 bit",
        0b10: "32 bit",
        0b11: "FLOAT",
    }

    def __init__(self):
        super().__init__()
        self.port = None
        self.mode = None
        self.info_type = None  # @see MsgPortModeInfoRequest
        self.value = None

    @classmethod
    def decode(cls, data):
        msg = super().decode(data)
        assert isinstance(msg, MsgPortModeInfo)
        msg.port = msg._byte()
        msg.mode = msg._byte()
        msg.info_type = msg._byte()
        msg.value = msg._value()
        return msg

    def _value(self):
        info = MsgPortModeInfoRequest
        if self.info_type == info.INFO_NAME:
            return self.payload[:self.payload.index(b"\00")].decode('ascii')
        elif self.info_type in (info.INFO_RAW_RANGE, info.INFO_PCT_RANGE, info.INFO_SI_RANGE):
            return [self._float(), self._float()]
        elif self.info_type == info.INFO_UNITS:
            return self.payload[:self.payload.index(b"\00")].decode('ascii')
        elif self.info_type == info.INFO_MAPPING:
            inp = self._bits_list(self._byte())
            outp = self._bits_list(self._byte())
            return {
                "input": [self.MAPPING_FLAGS[x] for x in inp],
                "output": [self.MAPPING_FLAGS[x] for x in outp],
            }
        elif self.info_type == info.INFO_MOTOR_BIAS:
            return self._byte()
        elif self.info_type == info.INFO_VALUE_FORMAT:
            return {
                "datasets": self._byte(),
                "type": self.DATASET_TYPES[self._byte()],
                "total_figures": self._byte(),
                "decimals": self._byte(),
            }
        else:
            return self.payload  # FIXME: will probably fail here


class MsgPortValueSingle(UpstreamMsg):
    """
    https://lego.github.io/lego-ble-wireless-protocol-docs/index.html#port-value-single
    """

    TYPE = 0x45

    def __init__(self):
        super().__init__()
        self.port = None

    @classmethod
    def decode(cls, data):
        msg = super().decode(data)
        assert isinstance(msg, MsgPortValueSingle)
        msg.port = msg._byte()
        return msg


class MsgPortValueCombined(UpstreamMsg):
    """
    https://lego.github.io/lego-ble-wireless-protocol-docs/index.html#port-value-combinedmode
    """

    TYPE = 0x46

    def __init__(self):
        super().__init__()
        self.port = None

    @classmethod
    def decode(cls, data):
        msg = super().decode(data)
        assert isinstance(msg, MsgPortValueCombined)
        msg.port = msg._byte()
        return msg


class MsgPortInputFmtSingle(UpstreamMsg):
    """
    https://lego.github.io/lego-ble-wireless-protocol-docs/index.html#port-input-format-single
    """

    TYPE = 0x47

    def __init__(self, port=None, mode=None, upd_enabled=None, upd_delta=None):
        super().__init__()
        self.port = port
        self.mode = mode
        self.upd_delta = upd_delta
        self.upd_enabled = upd_enabled

    @classmethod
    def decode(cls, data):
        msg = super().decode(data)
        assert isinstance(msg, MsgPortInputFmtSingle)
        msg.port = msg._byte()
        msg.mode = msg._byte()
        msg.upd_delta = msg._long()
        if len(msg.payload):
            msg.upd_enabled = msg._byte()

        return msg


class MsgPortInputFmtCombined(UpstreamMsg):  # TODO
    """
    https://lego.github.io/lego-ble-wireless-protocol-docs/index.html#port-input-format-combinedmode
    """

    TYPE = 0x48

    def __init__(self):
        super().__init__()
        self.port = None
        self.combined_control = None

    @classmethod
    def decode(cls, data):
        msg = super().decode(data)
        assert isinstance(msg, MsgPortInputFmtSingle)
        msg.port = msg._byte()
        return msg


class MsgVirtualPortSetup(DownstreamMsg):
    """
    https://lego.github.io/lego-ble-wireless-protocol-docs/index.html#virtual-port-setup
    """

    TYPE = 0x61

    CMD_DISCONNECT = 0x00
    CMD_CONNECT = 0x01

    def __init__(self, cmd, port):
        super().__init__()
        self.payload = pack("<B", cmd)
        if cmd == self.CMD_DISCONNECT:
            assert isinstance(port, int)
            self.payload += pack("<B", port)
        else:
            assert isinstance(port, (list, tuple))
            self.payload += pack("<B", port[0]) + pack("<B", port[1])


class MsgPortOutput(DownstreamMsg):
    """
    https://lego.github.io/lego-ble-wireless-protocol-docs/index.html#port-output-command
    """

    TYPE = 0x81

    SC_NO_BUFFER = 0b00000001
    SC_FEEDBACK = 0b00010000

    WRITE_DIRECT = 0x50
    WRITE_DIRECT_MODE_DATA = 0x51

    def __init__(self, port, subcommand, params, wait_complete=True):
        super().__init__()
        self.port = port
        self.is_buffered = False
        self.do_feedback = True
        self.wait_complete = wait_complete
        self.subcommand = subcommand
        self.params = params

    def bytes(self):
        startup_completion_flags = 0
        if not self.is_buffered:
            startup_completion_flags |= self.SC_NO_BUFFER

        if self.do_feedback:
            startup_completion_flags |= self.SC_FEEDBACK
            self.needs_reply = True

        self.payload = (
            pack("<B", self.port)
            + pack("<B", startup_completion_flags)
            + pack("<B", self.subcommand)
            + self.params
        )
        return super().bytes()

    def is_reply(self, msg):
        return (
            isinstance(msg, MsgPortOutputFeedback)
            and msg.port == self.port
            and (not self.wait_complete and msg.is_in_progress() or
                 msg.is_completed() or
                 self.is_buffered)
        )


class MsgPortOutputFeedback(UpstreamMsg):
    TYPE = 0x82

    def __init__(self):
        super().__init__()
        self.port = None
        self.status = None

    @classmethod
    def decode(cls, data):
        msg = super().decode(data)
        assert isinstance(msg, MsgPortOutputFeedback)
        assert len(msg.payload) == 2, "TODO: implement multi-port feedback message"
        msg.port = msg._byte()
        msg.status = msg._byte()
        return msg

    def is_in_progress(self):
        return self.status & 0b0001

    def is_completed(self):
        return self.status & 0b0010

    def is_discarded(self):
        return self.status & 0b0100

    def is_idle(self):
        return self.status & 0b1000


UPSTREAM_MSGS = (
    MsgHubProperties, MsgHubAction, MsgHubAlert, MsgHubAttachedIO, MsgGenericError,
    MsgPortInfo, MsgPortModeInfo,
    MsgPortValueSingle, MsgPortValueCombined, MsgPortInputFmtSingle, MsgPortInputFmtCombined,
    MsgPortOutputFeedback
)



from setuptools import setup
import os

version = "0.1.0"
tag = os.getenv("TRAVIS_TAG")
assert not tag or tag == version, "Travis tag has to match version: %r!=%r" % (tag, version)

setup(
    name="legoboost",
    version=version,

    author="Daniel Ersen",
    author_email="danielersen@gmail.com",
    license="MIT",
    description="Python library Lego Boost forked by the pylgbst repository but the accelerometer is a new function.",
    url='https://github.com/danielersen/Lego-Boost',
    keywords=['LEGO', 'ROBOTICS', 'BLUETOOTH'],

    packages=["pylgbst", "pylgbst.comms"],
    requires=[],
    extras_require={
        # Note that dbus and gi are normally system packages
        "gatt": ["gatt", "dbus", "gi"],
        "gattlib": ["gattlib"],
        "pygatt": ["pygatt", "pexpect"],
        "bluepy": ["bluepy"],
        "bleak": ["bleak"],
    },
)



import logging
import time
import unittest

from pylgbst.hub import MoveHub
from pylgbst.peripherals import LEDRGB, TiltSensor, COLOR_RED, Button, Current, Voltage, VisionSensor, \
    EncodedMotor
from tests import HubMock


class PeripheralsTest(unittest.TestCase):

    def test_button(self):
        hub = HubMock()
        time.sleep(0.1)
        button = Button(hub)
        hub.peripherals[0x00] = button

        vals = []

        def callback(pressed):
            vals.append(pressed)

        hub.connection.notification_delayed("060001020600", 0.0)
        button.subscribe(callback)
        time.sleep(0.1)

        hub.connection.notification_delayed("060001020600", 0.1)
        hub.connection.notification_delayed("060001020601", 0.2)
        hub.connection.notification_delayed("060001020600", 0.3)
        time.sleep(0.4)

        button.unsubscribe(callback)
        time.sleep(0.1)
        hub.connection.wait_notifications_handled()

        self.assertEqual([0, 1, 0], vals)
        self.assertEqual(b"0500010202", hub.writes[1][1])
        self.assertEqual(b"0500010203", hub.writes[2][1])

    def test_led(self):
        hub = HubMock()
        hub.led = LEDRGB(hub, MoveHub.PORT_LED)
        hub.peripherals[MoveHub.PORT_LED] = hub.led

        hub.connection.notification_delayed("0a004732000100000000", 0.1)
        hub.connection.notification_delayed("050082320a", 0.2)
        hub.led.set_color(COLOR_RED)
        self.assertEqual(b"0a004132000100000000", hub.writes.pop(1)[1])
        self.assertEqual(b"0800813211510009", hub.writes.pop(1)[1])

        hub.connection.notification_delayed("0a004732010100000000", 0.1)
        hub.connection.notification_delayed("050082320a", 0.2)
        hub.led.set_color((32, 64, 96))
        self.assertEqual(b"0a004132010100000000", hub.writes.pop(1)[1])
        self.assertEqual(b"0a008132115101204060", hub.writes.pop(1)[1])

    def test_current(self):
        hub = HubMock()
        time.sleep(0.1)
        current = Current(hub, MoveHub.PORT_CURRENT)
        hub.peripherals[MoveHub.PORT_CURRENT] = current

        vals = []

        def callback(val):
            vals.append(val)

        hub.connection.notification_delayed("0a00473b000100000001", 0.1)
        current.subscribe(callback)

        hub.connection.notification_delayed("0600453ba400", 0.1)
        time.sleep(0.2)

        hub.connection.notification_delayed("0a00473b000000000000", 0.1)
        current.unsubscribe(callback)
        hub.connection.wait_notifications_handled()

        self.assertEqual([97.87936507936507], vals)
        self.assertEqual(b"0a00413b000100000001", hub.writes[1][1])
        self.assertEqual(b"0a00413b000100000000", hub.writes[2][1])

    def test_voltage(self):
        hub = HubMock()
        time.sleep(0.1)
        voltage = Voltage(hub, MoveHub.PORT_VOLTAGE)
        hub.peripherals[MoveHub.PORT_VOLTAGE] = voltage

        vals = []

        def callback(val):
            vals.append(val)

        hub.connection.notification_delayed("0a00473c000100000001", 0.1)
        voltage.subscribe(callback)

        hub.connection.notification_delayed("0600453c9907", 0.1)
        time.sleep(0.2)

        hub.connection.notification_delayed("0a00473c000000000000", 0.1)
        voltage.unsubscribe(callback)
        hub.connection.wait_notifications_handled()

        self.assertEqual([4.79630105317236], vals)
        self.assertEqual(b"0a00413c000100000001", hub.writes[1][1])
        self.assertEqual(b"0a00413c000100000000", hub.writes[2][1])

    def test_tilt_sensor(self):
        hub = HubMock()
        sensor = TiltSensor(hub, MoveHub.PORT_TILT_SENSOR)
        hub.peripherals[MoveHub.PORT_TILT_SENSOR] = sensor

        hub.connection.notification_delayed('0a00473a000100000000', 0.05)
        hub.connection.notification_delayed('0600453a0201', 0.1)
        self.assertEqual((2, 1), sensor.get_sensor_data(TiltSensor.MODE_2AXIS_ANGLE))

        hub.connection.notification_delayed('0a00473a010100000000', 0.05)
        hub.connection.notification_delayed('0500453a00', 0.1)
        self.assertEqual((0,), sensor.get_sensor_data(TiltSensor.MODE_2AXIS_SIMPLE))

        hub.connection.notification_delayed('0a00473a020100000000', 0.05)
        hub.connection.notification_delayed('0600453a0201', 0.1)
        self.assertEqual((2,), sensor.get_sensor_data(TiltSensor.MODE_3AXIS_SIMPLE))

        hub.connection.notification_delayed('0a00473a030100000000', 0.05)
        hub.connection.notification_delayed('0800453a00000000', 0.1)
        self.assertEqual((0,), sensor.get_sensor_data(TiltSensor.MODE_IMPACT_COUNT))

        hub.connection.notification_delayed('0a00473a040100000000', 0.05)
        hub.connection.notification_delayed('0700453afd0140', 0.1)
        self.assertEqual((-3, 1, 64), sensor.get_sensor_data(TiltSensor.MODE_3AXIS_ACCEL))

        hub.connection.notification_delayed('0a00473a050100000000', 0.05)
        hub.connection.notification_delayed('0500453a00', 0.1)
        self.assertEqual((0,), sensor.get_sensor_data(TiltSensor.MODE_ORIENT_CF))

        hub.connection.notification_delayed('0a00473a060100000000', 0.05)
        hub.connection.notification_delayed('0600453a7f14', 0.1)
        self.assertEqual((127,), sensor.get_sensor_data(TiltSensor.MODE_IMPACT_CF))

        hub.connection.notification_delayed('0a00473a070100000000', 0.05)
        hub.connection.notification_delayed('0700453a00feff', 0.1)
        self.assertEqual((0, 254, 255), sensor.get_sensor_data(TiltSensor.MODE_CALIBRATION))

        hub.connection.wait_notifications_handled()

    def test_motor(self):
        hub = HubMock()
        motor = EncodedMotor(hub, MoveHub.PORT_D)
        hub.peripherals[MoveHub.PORT_D] = motor

        hub.connection.notification_delayed('050082030a', 0.1)
        motor.start_power(1.0)
        self.assertEqual(b"07008103110164", hub.writes.pop(1)[1])

        hub.connection.notification_delayed('050082030a', 0.1)
        motor.stop()
        self.assertEqual(b"0c0081031109000064647f03", hub.writes.pop(1)[1])

        hub.connection.notification_delayed('050082030a', 0.1)
        motor.set_acc_profile(1.0)
        self.assertEqual(b"090081031105e80300", hub.writes.pop(1)[1])

        hub.connection.notification_delayed('050082030a', 0.1)
        motor.set_dec_profile(1.0)
        self.assertEqual(b"090081031106e80300", hub.writes.pop(1)[1])

        hub.connection.notification_delayed('050082030a', 0.1)
        motor.start_speed(1.0)
        self.assertEqual(b"090081031107646403", hub.writes.pop(1)[1])

        hub.connection.notification_delayed('050082030a', 0.1)
        motor.stop()
        self.assertEqual(b"0c0081031109000064647f03", hub.writes.pop(1)[1])

        logging.debug("\n\n")
        hub.connection.notification_delayed('0500820301', 0.1)
        hub.connection.notification_delayed('050082030a', 0.2)
        motor.timed(1.0)
        self.assertEqual(b"0c0081031109e80364647f03", hub.writes.pop(1)[1])

        hub.connection.notification_delayed('0500820301', 0.1)
        hub.connection.notification_delayed('050082030a', 0.2)
        motor.angled(180)
        self.assertEqual(b"0e008103110bb400000064647f03", hub.writes.pop(1)[1])

        hub.connection.notification_delayed('050082030a', 0.2)
        motor.preset_encoder(-180)
        self.assertEqual(b"0b0081031151024cffffff", hub.writes.pop(1)[1])

        hub.connection.notification_delayed('0500820301', 0.1)
        hub.connection.notification_delayed('050082030a', 0.2)
        motor.goto_position(0)
        self.assertEqual(b"0e008103110d0000000064647f03", hub.writes.pop(1)[1])

        hub.connection.wait_notifications_handled()

    def test_motor_sensor(self):
        hub = HubMock()
        motor = EncodedMotor(hub, MoveHub.PORT_C)
        hub.peripherals[MoveHub.PORT_C] = motor

        vals = []

        def callback(*args):
            vals.append(args)

        hub.connection.notification_delayed('0a004702020100000001', 0.1)
        motor.subscribe(callback)

        hub.connection.notification_delayed("0800450200000000", 0.1)
        hub.connection.notification_delayed("08004502ffffffff", 0.2)
        hub.connection.notification_delayed("08004502feffffff", 0.3)
        time.sleep(0.4)

        hub.connection.notification_delayed('0a004702020000000000', 0.1)
        motor.unsubscribe(callback)
        hub.connection.wait_notifications_handled()

        self.assertEqual([(0,), (-1,), (-2,)], vals)

    def test_color_sensor(self):
        hub = HubMock()
        cds = VisionSensor(hub, MoveHub.PORT_C)
        hub.peripherals[MoveHub.PORT_C] = cds

        vals = []

        def callback(*args):
            vals.append(args)

        hub.connection.notification_delayed('0a00 4702080100000001', 0.1)
        cds.subscribe(callback)

        hub.connection.notification_delayed("08004502ff0aff00", 0.1)
        time.sleep(0.2)

        hub.connection.notification_delayed('0a00 4702090100000001', 0.1)
        cds.unsubscribe(callback)
        hub.connection.wait_notifications_handled()

        self.assertEqual([(255, 10.0)], vals)



import logging

import pygatt

from pylgbst.comms import Connection, MOVE_HUB_HW_UUID_CHAR
from pylgbst.utilities import str2hex

log = logging.getLogger('comms-pygatt')


class GattoolConnection(Connection):
    """
    Used for connecting to

    :type _conn_hnd: pygatt.backends.bgapi.device.BGAPIBLEDevice
    """

    def __init__(self, controller='hci0'):
        Connection.__init__(self)
        self.backend = lambda: pygatt.GATTToolBackend(hci_device=controller)
        self._conn_hnd = None

    def connect(self, hub_mac=None, hub_name=None):
        log.debug("Trying to connect client to MoveHub with MAC: %s", hub_mac)
        adapter = self.backend()
        adapter.start()  # enable or disable restart? What's the best?

        while not self._conn_hnd:
            log.info("Discovering devices...")
            devices = adapter.scan(1)
            log.debug("Devices: %s", devices)

            # Pass each device found to _is_device_matched( ) to see if it the device we want
            for dev in devices:
                address = dev['address']
                name = dev['name']
                if self._is_device_matched(address, name, hub_mac, hub_name):
                    self._conn_hnd = adapter.connect(address)
                    break

            if self._conn_hnd:
                break

        return self

    def disconnect(self):
        self._conn_hnd.disconnect()

    def write(self, handle, data):
        log.debug("Writing to handle %s: %s", handle, str2hex(data))
        return self._conn_hnd.char_write_handle(handle, bytearray(data))

    def set_notify_handler(self, handler):
        self._conn_hnd.subscribe(MOVE_HUB_HW_UUID_CHAR, handler)

    def is_alive(self):
        return True


class BlueGigaConnection(GattoolConnection):
    def __init__(self):
        super().__init__()
        self.backend = lambda: pygatt.BGAPIBackend()



import sys
import time
import unittest

import bleak
from packaging import version

import pylgbst
import pylgbst.comms.cbleak as cbleak

bleak.BleakClient = object()
bleak.discover = object()

last_response = None
lt37 = version.parse(sys.version.split(' ')[0]) < version.parse("3.7")


class BleakDriverTest(unittest.TestCase):
    def test_driver_creation(self):
        connection = pylgbst.get_connection_bleak()
        self.assertIsInstance(connection, cbleak.BleakDriver)
        self.assertFalse(connection.is_alive(), 'Checking that factory returns not started driver')

    @unittest.skipIf(lt37, "Python version is too low")
    def test_communication(self):
        driver = cbleak.BleakDriver()

        async def fake_thread():
            print('Fake thread initialized')
            while not driver._abort:
                time.sleep(0.1)
                if driver.req_queue.qsize() != 0:
                    print('Received data, sending back')
                    data = driver.req_queue.get()
                    driver.resp_queue.put(data)

        driver._bleak_thread = fake_thread
        driver.set_notify_handler(BleakDriverTest.validation_handler)
        driver.enable_notifications()

        time.sleep(0.5)  # time for driver initialization
        self.assertTrue(driver.is_alive(), 'Checking that driver starts')
        handle = 0x32
        data = [0xD, 0xE, 0xA, 0xD, 0xB, 0xE, 0xE, 0xF]
        driver.write(handle, data)
        time.sleep(0.5)  # processing time
        self.assertEqual(handle, last_response[0], 'Verifying response handle')
        self.assertEqual(bytes(data), last_response[1], 'Verifying response data')

        driver.disconnect()
        time.sleep(0.5)  # processing time
        self.assertFalse(driver.is_alive())

    @staticmethod
    def validation_handler(handle, data):
        global last_response
        last_response = (handle, data)


if __name__ == '__main__':
    unittest.main()



"""
This module offers some utilities, in a way they are work in both Python 2 and 3
"""

import binascii
import logging
import math
import sys
from struct import unpack

log = logging.getLogger(__name__)

if sys.version_info[0] == 2:
    import Queue as queue
else:
    import queue as queue

queue = queue  # just to use it


def check_unpack(seq, index, pattern, size):
    """Check that we got size bytes, if so, unpack using pattern"""
    data = seq[index: index + size]
    assert len(data) == size, "Unexpected data len %d, expected %d" % (len(data), size)
    return unpack(pattern, data)[0]


def usbyte(seq, index):
    return check_unpack(seq, index, "<B", 1)


def ushort(seq, index):
    return check_unpack(seq, index, "<H", 2)


def usint(seq, index):
    return check_unpack(seq, index, "<I", 4)


def str2hex(data):  # we need it for python 2+3 compatibility
    # if sys.version_info[0] == 3:
    # data = bytes(data, 'ascii')
    if not isinstance(data, (bytes, bytearray)):
        data = bytes(data, "ascii")
    hexed = binascii.hexlify(data)
    return hexed


def abs_scaled_100(relative):
    if relative < -1.0:
        log.warning("Speed cannot be less than -1")
        relative = -1.0

    if relative > 1.0:
        log.warning("Speed cannot be more than 1")
        relative = 1.0

    absolute = math.ceil(relative * 100)  # scale of 100 is proven by experiments
    return int(absolute)
